/*
 * FTMEXT SDK example
 * Copyright (C) 2004 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include "ftmlib.h"

typedef struct
{
  ftmext_t o; /* mandatory FTM external "super class" */
  double num;
  fts_symbol_t sym;
  fts_object_t *obj;
  fts_atom_t val;
} coucou_t;

/************************************************
 *
 *  output methods
 *
 */
FTMEXT_METHOD_VOID(coucou, output)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  fts_atom_t a[4];
  
  fts_set_float(a, self->num);
  fts_set_symbol(a + 1, self->sym);
  fts_set_object(a + 2, self->obj);
  a[3] = self->val;
  
  ftmext_outlet_varargs((ftmext_t *)self, 0, 4, a);
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_VOID(coucou, print)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  fts_atom_t a_obj;

  fts_post("couocu!\n");

  fts_post("  number: %g\n", self->num);
  fts_post("  symbol: %s\n", fts_symbol_name(self->sym));

  fts_set_object(&a_obj, self->obj);
  fts_post("  object: ");
  fts_post_atoms(1, &a_obj);
  fts_post("\n");
  
  fts_post("  arbitrary value: ");
  fts_post_atoms(1, &self->val);
  fts_post("\n");
  
  FTMEXT_METHOD_RETURN;
}

/****************************************************
 *
 *  set methods
 *
 */
FTMEXT_METHOD_NUMBER(coucou, num)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  double num = FTMEXT_GET_FLOAT();
  
  if(num != 7.)
	self->num = num;
  else
	ftmext_error((ftmext_t *)self, "doesn't like 7");
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_SYMBOL(coucou, sym)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  fts_symbol_t sym = FTMEXT_GET_SYMBOL();
  
  if(sym != NULL)
  {
	if(sym != fts_new_symbol("seven"))
	  self->sym = sym;
	else
	  ftmext_error((ftmext_t *)self, "doesn't like 'seven'");
  }
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_OBJECT(coucou, NULL, obj)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  fts_object_t *obj = FTMEXT_GET_OBJECT();
  
  if(obj != NULL)
  {
	if(fts_object_is_a(obj, fts_tuple_class))
	{
	  int size = fts_tuple_get_size((fts_tuple_t *)obj);
	  fts_atom_t *atoms = fts_tuple_get_atoms((fts_tuple_t *)obj);
	  
	  if(size == 7 || (fts_is_number(atoms) && fts_get_number_int(atoms) == 7))
	  {
		ftmext_error((ftmext_t *)self, "doesn't like tuples of 7 elements nor tuples starting with 7");
		FTMEXT_EXIT;
	  }
	}
	  
	if(self->obj != NULL)
	  fts_object_release(self->obj);
	
	self->obj = obj;
	fts_object_refer(obj);
  }
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_ANY(coucou, val)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  fts_atom_t *val = FTMEXT_GET_ANY();
  
  if(!fts_is_void(val))
  {
	/* copy the incoming value into the object and put a reference when needed (FTM object references) */
	fts_atom_assign(&self->val, val);
	
	/* same as:
	self->val = *val;
	fts_atom_refer(val); */
  }
  
  FTMEXT_METHOD_RETURN;
}

/****************************************************
 *
 *  get methods
 *
 */
FTMEXT_METHOD_VOID(coucou, getnum)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  
  ftmext_outlet_number((ftmext_t *)self, 0, self->num);
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_VOID(coucou, getsym)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  
  ftmext_outlet_symbol((ftmext_t *)self, 0, self->sym);
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_VOID(coucou, getobj)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  
  ftmext_outlet_object((ftmext_t *)self, 0, self->obj);
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_VOID(coucou, getval)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  
  ftmext_outlet_any((ftmext_t *)self, 0, &self->val);
  
  FTMEXT_METHOD_RETURN;
}

/****************************************************
 *
 *  additional set methods
 *
 */
FTMEXT_METHOD_NUMBER(coucou, setnum)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();

  coucou_num((ftmext_t *)self, FTMEXT_GET_NUMBER());
	
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_SYMBOL(coucou, setsym)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();

  coucou_sym((ftmext_t *)self, FTMEXT_GET_SYMBOL());
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_OBJECT(coucou, NULL, setobj)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();

  coucou_obj((ftmext_t *)self, FTMEXT_GET_OBJECT());
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_ANY(coucou, setval)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  
  coucou_val((ftmext_t *)self, FTMEXT_GET_ANY());
  
  FTMEXT_METHOD_RETURN;
}

/****************************************************
 *
 *  inlet method
 *
 */
FTMEXT_INLET(coucou)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  int inlet = FTMEXT_GET_INLET();

  switch(inlet)
  {
	case 0:
	  ftmext_error((ftmext_t *)self, "this cannot happen");
	  break;
	case 1:
	  coucou_num((ftmext_t *)self, FTMEXT_GET_NUMBER());
	  break;
	case 2:
	  coucou_sym((ftmext_t *)self, FTMEXT_GET_SYMBOL());
	  break;
	case 3:
	  coucou_obj((ftmext_t *)self, FTMEXT_GET_OBJECT());
	  break;
	case 4:
	  coucou_val((ftmext_t *)self, FTMEXT_GET_ANY());
	  break;
  }
  
  FTMEXT_INLET_RETURN;  
}

/****************************************************
 *
 *  class
 *
 */

/* constructor */
FTMEXT_INIT(coucou)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();
  
  self->num = 0.0;
  self->sym = fts_new_symbol("coucou");
  
  fts_set_int(&self->val, 0);
    
  self->obj = fts_object_create(fts_tuple_class, 1, &self->val);
  fts_object_refer(self->obj);
  
  FTMEXT_INIT_RETURN;
}

/* deconstructor */
FTMEXT_DELETE(coucou)
{
  coucou_t *self = (coucou_t *)FTMEXT_GET_EXT();

  fts_object_release(self->obj);
  fts_atom_release(&self->val);

  FTMEXT_DELETE_RETURN;
}

/* class definition */
FTMEXT_CLASS(fff, coucou, coucou_t)
{
  FTMEXT_CLASS_DESCR(_prefix_ extclass, "coucou", "\
French: medium-sized long-tailed bird, typically with a gray or brown back and barred or pale underparts");

  FTMEXT_CLASS_ARGUMENT_NUMBER(coucou, 0, num, 0.0, "<num: a number> - set the number");
  FTMEXT_CLASS_ARGUMENT_SYMBOL(coucou, 1, sym, NULL, "<sym: a symbol> - set the symbol");
  FTMEXT_CLASS_ARGUMENT_OBJECT(coucou, NULL, 2, obj, NULL, "<obj: an object> - set the object");
  FTMEXT_CLASS_ARGUMENT_ANY(coucou, 3, val, fts_null, "<any: any value> - set the arbitrary value");
  
  FTMEXT_CLASS_ATTRIBUTE_NUMBER(coucou, num, FTMEXT_ATTRIBUTE_DYNAMIC, "<num: a number> - set the number");
  FTMEXT_CLASS_ATTRIBUTE_SYMBOL(coucou, sym, FTMEXT_ATTRIBUTE_DYNAMIC, "<sym: a symbol> - set the symbol");
  FTMEXT_CLASS_ATTRIBUTE_OBJECT(coucou, NULL, obj, FTMEXT_ATTRIBUTE_DYNAMIC, "<obj: an object> - set the object");
  FTMEXT_CLASS_ATTRIBUTE_ANY(coucou, val, FTMEXT_ATTRIBUTE_DYNAMIC, "<any: an arbitrary value> - set the arbitrary value");

  FTMEXT_CLASS_MESSAGE_BANG(coucou, output, "- output all stored values");
  FTMEXT_CLASS_MESSAGE_VOID(coucou, print, "- print all stored values");

  FTMEXT_CLASS_MESSAGE_VOID(coucou, getnum, "- output the number");
  FTMEXT_CLASS_MESSAGE_VOID(coucou, getsym, "- output the symbol");
  FTMEXT_CLASS_MESSAGE_VOID(coucou, getobj, "- output the object");
  FTMEXT_CLASS_MESSAGE_VOID(coucou, getval, "- output the arbitrary value");

  FTMEXT_CLASS_MESSAGE_NUMBER(coucou, setnum, "<num: a number> - compare a number to the stored number");
  FTMEXT_CLASS_MESSAGE_SYMBOL(coucou, setsym, "<sym: a symbol> - set the symbol");
  FTMEXT_CLASS_MESSAGE_OBJECT(coucou, NULL, setobj, "<obj: an object> - set the object");
  FTMEXT_CLASS_MESSAGE_ANY(coucou, setval, "<any: any value> - set the object");
  
  FTMEXT_CLASS_INLET_NUMBER(coucou, 1, "<num: a number> - set the number");
  FTMEXT_CLASS_INLET_SYMBOL(coucou, 2, "<sym: a symbol> - set the symbol");
  FTMEXT_CLASS_INLET_OBJECT(coucou, NULL, 3, "<obj: an object> - set the object");
  FTMEXT_CLASS_INLET_ANY(coucou, 4, "<any: an arbitrary value> - set the arbitrary value");

  FTMEXT_CLASS_OUTLET_VARARGS(coucou, 0, "output stored values");

  FTMEXT_CLASS_RETURN;
}
