/*
 * FTM
 * Copyright (C) 2004 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

/**
 *  atom and helper functions
 *
 * @defgroup fts_atom FTS atom (value of arbitrary type)
 * @ingroup fts_system
 */

#ifndef _FTS_ATOM_H
#define _FTS_ATOM_H

#include "ftmtypes.h"

/* primitive atom pseudo classes */
FTS_API fts_class_t *fts_void_class;
FTS_API fts_class_t *fts_number_class;
FTS_API fts_class_t *fts_symbol_class;
FTS_API fts_class_t *fts_pointer_class;
FTS_API fts_class_t *fts_string_class;

/* platform dependent implementations of FTS words */
#define fts_word_set_int ftmrte_word_set_int
#define fts_word_set_float ftmrte_word_set_float
#define fts_word_set_symbol ftmrte_word_set_symbol
#define fts_word_set_object ftmrte_word_set_object
#define fts_word_set_pointer ftmrte_word_set_pointer
#define fts_word_set_string ftmrte_word_set_string

#define fts_word_get_int ftmrte_word_get_int
#define fts_word_get_float ftmrte_word_get_float
#define fts_word_get_symbol ftmrte_word_get_symbol
#define fts_word_get_object ftmrte_word_get_object
#define fts_word_get_pointer ftmrte_word_get_pointer
#define fts_word_get_string ftmrte_word_get_string

#ifdef AVOID_MACROS

/** @name Set atom values
*  @{ */

/**
* @fn void fts_set_void(fts_atom_t *atom) 
 * @brief set atom value to void
 * @param atom pointer to the atom
 * @ingroup fts_atom
 */
void fts_set_void(fts_atom_t *atom);

/**
* @fn void fts_set_int(fts_atom_t *atom, int value) 
 * @brief set atom value to given int 
 * @param atom pointer to the atom
 * @param value int value
 * @ingroup fts_atom
 */
void fts_set_int(fts_atom_t *atom, int value);

/**
* @fn void fts_set_float(fts_atom_t *atom, float value)
 * @brief set atom value to given float 
 * @param atom pointer to the atom
 * @param value float value
 * @ingroup fts_atom
 */
void fts_set_float(fts_atom_t *atom, float value);

/**
* @fn void fts_set_symbol(fts_atom_t *atom, fts_symbol_t sym)
 * @brief set atom value to given symbol 
 * @param atom pointer to the atom
 * @param sym symbol value
 * @ingroup fts_atom
 */
void fts_set_symbol(fts_atom_t *atom, fts_symbol_t sym);

/**
* @fn void fts_set_object(fts_atom_t *atom, fts_object_t *obj)
 * @brief set atom value to given FTS object (reference) 
 * @param atom pointer to the atom
 * @param obj FTS object value
 * @ingroup fts_atom
 */
void fts_set_object(fts_atom_t *atom, fts_object_t *obj);

/**
* @fn void fts_set_pointer(fts_atom_t *atom, void *ptr) 
 * @brief set atom value to given pointer 
 * @param atom pointer to the atom
 * @param ptr pointer value
 * @ingroup fts_atom
 */
void fts_set_pointer(fts_atom_t *atom, void *ptr);

/**
* @fn void fts_set_string(fts_atom_t *atom, char *string)
 * @brief set atom value to given string 
 * @param atom pointer to the atom
 * @param string string value
 * @ingroup fts_atom
 */
void fts_set_string(fts_atom_t *atom, char *string);

/** @} */

#else

#define fts_set_void ftmrte_set_void
#define fts_set_int ftmrte_set_int
#define fts_set_float ftmrte_set_float
#define fts_set_symbol ftmrte_set_symbol
#define fts_set_object ftmrte_set_object
#define fts_set_pointer ftmrte_set_pointer
#define fts_set_string ftmrte_set_string

#endif

#ifdef AVOID_MACROS

/** @name Get atom values
*  @{ */

/**
* @fn int fts_get_int(const fts_atom_t *atom) 
 * @brief get int from given atom
 * @param atom pointer to the atom
 * @return int value
 * @ingroup fts_atom
 */
int fts_get_int(const fts_atom_t *atom);
/**
* @fn float fts_get_float(const fts_atom_t *atom) 
 * @brief get float from given atom
 * @param atom pointer to the atom
 * @return float value
 * @ingroup fts_atom
 */
float fts_get_float(const fts_atom_t *atom);
/**
* @fn int fts_get_number_int(const fts_atom_t *atom) 
 * @brief get int from given atom
 * @param atom pointer to the atom
 * @return int value
 * @ingroup fts_atom
 */
int fts_get_number_int(const fts_atom_t *atom);
/**
* @fn float fts_get_number_float(const fts_atom_t *atom) 
 * @brief get float from given atom
 * @param atom pointer to the atom
 * @return float value
 * @ingroup fts_atom
 */
float fts_get_number_float(const fts_atom_t *atom);
/**
* @fn fts_symbol_t fts_get_symbol(const fts_atom_t *atom) 
 * @brief get symbol from given atom 
 * @param atom pointer to the atom
 * @return symbol value
 * @ingroup fts_atom
 */
fts_symbol_t fts_get_symbol(const fts_atom_t *atom);
/**
* @fn fts_object_t *fts_get_object(const fts_atom_t *atom) 
 * @brief get object from given atom
 * @param atom pointer to the atom
 * @return FTS object reference (pointer)
 * @ingroup fts_atom
 */
fts_object_t *fts_get_object(const fts_atom_t *atom);

/**
* @fn void *fts_get_pointer(const fts_atom_t *atom) 
 * @brief get pointer from given atom
 * @param atom pointer to the atom
 * @return pointer value as void *
 * @ingroup fts_atom
 */
void *fts_get_pointer(const fts_atom_t *atom);

/**
* @fn char *fts_get_string(const fts_atom_t *atom) 
 * @brief get string from given atom
 * @param atom pointer to the atom
 * @return string value
 * @ingroup fts_atom
 */
char *fts_get_string(const fts_atom_t *atom);

/** @} */

#else

#define fts_get_int ftmrte_get_int
#define fts_get_float ftmrte_get_float
#define fts_get_number_int ftmrte_get_number_int
#define fts_get_number_float ftmrte_get_number_float
#define fts_get_symbol ftmrte_get_symbol
#define fts_get_object ftmrte_get_object
#define fts_get_pointer ftmrte_get_pointer
#define fts_get_string ftmrte_get_string

#endif

#ifdef AVOID_MACROS

/** @name Get and test atom type/class
*  @{ */

/**
* @fn int fts_is_void(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a void value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_void(const fts_atom_t *atom);

/**
* @fn int fts_is_int(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a int value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_int(const fts_atom_t *atom);

/**
* @fn int fts_is_float(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a float value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_float(const fts_atom_t *atom);

/**
* @fn int fts_is_number(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a number value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_number(const fts_atom_t *atom);

/**
* @fn int fts_is_symbol(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a symbol value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_symbol(const fts_atom_t *atom);

/**
* @def int fts_is_object(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains an fts_object value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_object(const fts_atom_t *atom);

/**
* @fn int fts_is_pointer(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a pointer value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_pointer(const fts_atom_t *atom);

/**
* @fn int fts_is_string(const fts_atom_t *atom) 
 * @brief tests whether the given atom contains a string value
 * @param atom pointer to the atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
int fts_is_string(const fts_atom_t *atom);

/**
* @fn fts_class_t *fts_get_class(const fts_atom_t *atom)
 * @brief get FTS class of the atom value
 * @param atom pointer to the atom
 * @return pointer to an FTS class
 * @ingroup fts_atom
 */
fts_class_t *fts_get_class(const fts_atom_t *atom);

/**
* @fn fts_symbol_t fts_get_class_name(const fts_atom_t *atom)
 * @brief get FTS class name of the atom value
 * @param atom pointer to the atom
 * @return class name
 * @ingroup fts_atom
 */
FTM_API fts_symbol_t fts_get_class_name(const fts_atom_t *atom);

/**
* @fn int fts_get_class_id(const fts_atom_t *atom)
 * @brief get FTS class id of the atom value
 * @param atom pointer to the atom
 * @return FTS class id
 * @ingroup fts_atom
 */
FTM_API int fts_get_class_id(const fts_atom_t *atom);

/**
* @fn int fts_is_a(const fts_atom_t *atom, fts_class_t *type)
 * @brief tests whether the atom value is of the given type (FTS class)
 * @param atom pointer to the atom
 * @param type pointer to FTS class
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
FTM_API int fts_is_a(const fts_atom_t *atom, fts_class_t *type);

/** @} */

#else

#define fts_is_void ftmrte_is_void
#define fts_is_int ftmrte_is_int
#define fts_is_float ftmrte_is_float
#define fts_is_number ftmrte_is_number
#define fts_is_symbol ftmrte_is_symbol
#define fts_is_object ftmrte_is_object
#define fts_is_pointer ftmrte_is_pointer
#define fts_is_string ftmrte_is_string

#define fts_get_class ftmrte_atom_get_class
#define fts_get_class_name(p) (fts_class_get_name(fts_get_class((p))))
#define fts_get_class_id(p) (ftmrte_atom_get_class_id(p))
#define fts_is_a(p,c) (fts_get_class(p) == (c))

#endif

#ifdef AVOID_MACROS

/** @name Misc atom functions
*  @{ */

/**
* Sets atom value to void (releases FTS object reference).
 *
 * @fn void fts_atom_void(const fts_atom_t *p)
 * @brief sets atom value to void (releases FTS object reference)
 * @param p atom pointer
 * @ingroup fts_atom
 */
FTM_API void fts_atom_void(const fts_atom_t *p);

/**
* Adds a reference to an FTS object contained by the atom.
 * Has no effect for primitive atom values.
 *
 * @fn void fts_atom_refer(const fts_atom_t *atom)
 * @brief adds reference to atom of FTS object reference
 * @param atom atom pointer
 * @ingroup fts_atom
 */
FTM_API void fts_atom_refer(const fts_atom_t *atom);

/**
* Removes a reference from an FTS object contained by the atom.
 * Has no effect for primitive atom values.
 *
 * @fn  void fts_atom_release(const fts_atom_t *atom)
 * @brief release reference to atom of FTS object reference
 * @param atom atom pointer
 * @ingroup fts_atom
 */
FTM_API void fts_atom_release(const fts_atom_t *atom);

#else

#define fts_atom_void(p) \
do \
{ \
  fts_atom_release(p); \
    fts_set_void(p); \
} \
while(0)

#define fts_atom_refer(p) do {if(fts_is_object(p)) fts_object_refer(fts_get_object(p));} while(0)
#define fts_atom_release(p) do {if(fts_is_object(p)) fts_object_release(fts_get_object(p));} while(0)

#endif

#ifdef AVOID_MACROS

/**
* Assigns the value of one atom to another.
 * For atoms containing FTS objects only the reference is copied.
 * The function properly handles references to FTS objects and requires initialized atoms.
 *
 * @fn void fts_atom_assign(const fts_atom_t *dest, const fts_atom_t *src)
 * @brief assign src atom value to dest atom (handles references to FTS object)
 * @param dest destination atom pointer
 * @param src source atom pointer
 * @ingroup fts_atom
 */
FTM_API void fts_atom_assign(const fts_atom_t *dest, const fts_atom_t *src);

#else

#define fts_atom_assign(dest, src) \
do \
{ \
  fts_atom_release(dest); \
    *(dest) = *(src); \
      fts_atom_refer(src); \
} \
while(0)

#endif

/**
* Copies atom value or object (not just reference) from one atom to another.
 * For FTM objects the copy function of the class is used.
 * The function properly handles references to FTS objects and requires initialized atoms.
 *
 * @fn void fts_atom_copy(const fts_atom_t *from, fts_atom_t *to)
 * @brief copies atom value to another
 * @param from pointer to original atom
 * @param to   pointer to destination atom
 * @ingroup fts_atom
 */
FTS_API void fts_atom_copy( const fts_atom_t *from, fts_atom_t *to);

#ifdef AVOID_MACROS

/**
* Tests whether two atoms values are of the same type (class).
 *
 * @fn int fts_atom_same_type(const fts_atom_t *atom1, const fts_atom_t *atom2)
 * @brief tests whether two atoms values are of the same type
 * @param atom1 pointer to an atom
 * @param atom2 another pointer to an atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
FTM_API int fts_atom_same_type(const fts_atom_t *atom1, const fts_atom_t *atom2);

#else

#define fts_atom_same_type(p1, p2) ((p1)->a_type == (p2)->a_type)

#endif

/**
* Tests whether two atoms values are identical. 
 * For atoms containing FTS objects only the references are compared.
 *
 * @fn int fts_atom_identical(const fts_atom_t *atom1, const fts_atom_t *atom2)
 * @brief tests whether two atoms values are identical (compares FTS object references)
 * @param atom1 first atom
 * @param atom2 second atom
 * @return boolean (0|1)
 * @ingroup fts_atom
 */
FTS_API int fts_atom_identical(const fts_atom_t *atom1, const fts_atom_t *atom2);

/**
* Tests whether two atoms are equal. 
 * For atoms containing FTS objects the class dependent comparison function is used.
 *
 * @fn int fts_atom_equals(const fts_atom_t *atom1, const fts_atom_t *atom2)
 * @brief tests whether two atoms have equal values (compares numbers and FTS object values)
 * @param atom1 first atom
 * @param atom2 second atom
 * @return 1 if equals
 * @ingroup fts_atom
 */
FTS_API int fts_atom_equals(const fts_atom_t *atom1, const fts_atom_t *atom2);

/**
* Compares atoms of number, symbol, and FTS objects (compares the references).
 *
 * @fn int fts_atom_compare(const fts_atom_t *atom1, const fts_atom_t *atom2)
 * @brief compares two atoms
 * @param atom1 pointer to atom
 * @param atom2 pointer to atom
 * @return a value < 0, 0, > 0 if atom1 < atom2, atom1 == atom2, atom1 > atom2 respectively
 * @ingroup fts_atom
 */
FTS_API int fts_atom_compare(const fts_atom_t *atom1, const fts_atom_t *atom2);

/** @} */

/**
* @var fts_atom_t *fts_null
 * @brief constant representing a 'void' atom
 * @ingroup fts_atom
 */
FTS_API fts_atom_t *fts_null;

/**
* @var fts_atom_t *fts_zero
 * @brief constant representing an int 0 (zero) atom
 * @ingroup fts_atom
 */
FTS_API fts_atom_t *fts_zero;

/**
* Constant representing a useless dummy atom for various occasions (might change to any value)
 * @var fts_atom_t *fts_nix
 * @brief dummy atom
 * @ingroup fts_atom
 */
FTS_API fts_atom_t *fts_nix;

#endif
