/*
 * FTM
 * Copyright (C) 2004 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

/**
* @defgroup ftm_externals FTM externals
*/

#ifndef _FTMEXT_H
#define _FTMEXT_H

#ifdef WIN32
#if defined(FTM_EXPORTS)
#define FTMEXT_API __declspec(dllexport)
#else
#define FTMEXT_API __declspec(dllimport)
#endif
#else
#define FTMEXT_API extern
#endif

#include "fts.h"
#include "ftmtypes.h"

/******************************************************************************
 *
 *  predefined symbols and types
 *
 */

typedef struct _ftmext_class_ ftmext_class_t;
typedef struct _ftmext_ ftmext_t;

FTMEXT_API char ftm_str_obj[];
FTMEXT_API char ftm_str_redefine[];

FTMEXT_API fts_symbol_t ftm_sym_obj;
FTMEXT_API fts_symbol_t ftm_sym_redefine;

typedef struct _ftmext_argument_ ftmext_argument_t;
typedef struct _ftmext_attribute_ ftmext_attribute_t;
typedef struct _ftmext_message_ ftmext_message_t;
typedef struct _ftmext_inlet_ ftmext_inlet_t;
typedef struct _ftmext_outlet_ ftmext_outlet_t;

/* pseudo classes used in method declarations */
#define FTMEXT_TYPEID_ANY 6
#define FTMEXT_TYPEID_VARARGS 7

FTMEXT_API fts_class_t *ftmext_pseudoclass_undefined;
FTMEXT_API fts_class_t *ftmext_pseudoclass_any;
FTMEXT_API fts_class_t *ftmext_pseudoclass_varargs;
FTMEXT_API fts_class_t *ftmext_pseudoclass_signal;
FTMEXT_API fts_class_t *ftmext_pseudoclass_anything;

/* method types */
typedef void (*ftmext_method_void_t)(ftmext_t *o);
typedef void (*ftmext_method_number_t)(ftmext_t *o, double f);
typedef void (*ftmext_method_symbol_t)(ftmext_t *o, fts_symbol_t s);
typedef void (*ftmext_method_object_t)(ftmext_t *o, fts_object_t *obj);
typedef void (*ftmext_method_any_t)(ftmext_t *o, fts_atom_t *a);
typedef void (*ftmext_method_varargs_t)(ftmext_t *o, int n, fts_atom_t *a);

typedef void (*ftmext_save_handler_t)(ftmext_t *ext);

/******************************************************************************
 *
 *  FTM pseudo package data
 *
 */
typedef void ftmext_package_data_init_t(void *);
FTMEXT_API void *ftmext_package_data(fts_symbol_t name, ftmext_package_data_init_t init, int size);

/******************************************************************************
 *
 *  FTM external class
 *
 */
struct _ftmext_class_
{
  fts_symbol_t name; /**< class name */
  
  char *descr_brief; /* brief description (a few words) */
  char *descr_full; /* full description (HTML) */
  
  /* inlets */
  int n_inlets_dsp; /**< number of DSP inlets */
  int n_inlets_ctl; /**< number of ctl inlets */
  int max_inlets; /**< max_inlets */
  ftmext_inlet_t *inlets; /**< inlets */
  int inlet_varargs_dispatch;
  
  /* outlets */
  int n_outlets_dsp; /**< n_outlets_dsp */
  int n_outlets_ctl; /**< n_outlets_ctl */
  int max_outlets; /**< max_outlets */
  ftmext_outlet_t *outlets; /**< outlets */
  
  /* arguments */
  ftmext_argument_t *arguments; /**< arguments */
  int n_arguments; /**< n_arguments */
  int max_arguments; /**< max_arguments */
  
  /* attributes and messages */
  ftmext_attribute_t *attributes; /**< attributes */
  ftmext_message_t *messages; /**< messages */
  
  /* alarm callback */
  ftmext_method_void_t alarm_callback;
  
  /* save callback */
  ftmext_save_handler_t save_handler;
  
  /* FTM runtime environment user data */
  void *handle; /**< FTM runtime environment user data */
};

#define ftmext_class_get_inlets(c) (((c)->n_inlets_dsp > (c)->n_inlets_ctl)? (c)->n_inlets_dsp: (c)->n_inlets_ctl)
#define ftmext_class_get_outlets(c) (((c)->n_outlets_dsp > (c)->n_outlets_ctl)? (c)->n_outlets_dsp: (c)->n_outlets_ctl)

#define ftmext_class_get_name(c) ((c)->name)
#define ftmext_class_is_dsp(c) ((c)->n_inlets_dsp + (c)->n_outlets_dsp > 0)

#define ftmext_class_set_alarm_callback(c, x) ((c)->alarm_callback = (x))
#define ftmext_class_get_alarm_callback(c) ((c)->alarm_callback)

FTMEXT_API ftmext_class_t *ftmext_class_get_by_name(fts_symbol_t name);
FTMEXT_API ftmext_class_t *ftmext_class_new(fts_symbol_t name);
FTMEXT_API void ftmext_class_descr(ftmext_class_t *cl, char *brief, char *full);

/******************************************************************************
 *
 *  FTM external module
 *
 */
struct _ftmext_
{
  ftmrte_external_head_t head; /**< head ... */
  ftmext_class_t *cl; /**< class  ... */
  fts_patcher_t *scope; /**< scope ... */
  fts_expression_t *expression;  /**< expression ... */
  fts_status_t expression_status; /**< expression_status ... */
  fts_array_t arguments; /**< arguments ... */
  int ref_init; /**< ref_init ... */
  int arg_init; /**< arg_init ... */
  int attr_init; /**< attr_init ... */
  int n_inlets_ctl; /**< n_inlets_ctl ... */
  int n_outlets_ctl; /**< n_outlets_ctl ... */
  int n_inlets_dsp; /**< n_inlets_dsp ... */
  int n_outlets_dsp; /**< n_outlets_dsp ... */
  int *dsp_connected;
};

#define ftmext_get_class(e) ((e)->cl)
#define ftmext_get_class_name(e) ftmext_class_get_name(ftmext_get_class(e))
#define ftmext_get_scope(e) ((e)->scope)
#define ftmext_set_scope(e, s) ((e)->scope = (s))
#define ftmext_get_expression_status(e) ((e)->expression_status)
#define ftmext_get_inlets_ctl(e) ((e)->n_inlets_ctl)
#define ftmext_get_inlets_dsp(e) ((e)->n_inlets_dsp)
#define ftmext_get_outlets_ctl(e) ((e)->n_outlets_ctl)
#define ftmext_get_outlets_dsp(e) ((e)->n_outlets_dsp)
#define ftmext_get_inlets(e) (((e)->n_inlets_dsp > (e)->n_inlets_ctl)? (e)->n_inlets_dsp: (e)->n_inlets_ctl)
#define ftmext_get_outlets(e) (((e)->n_outlets_dsp > (e)->n_outlets_ctl)? (e)->n_outlets_dsp: (e)->n_outlets_ctl)

#define ftmext_get_connected(e, i) ((e)->dsp_connected[((i) < (e)->n_inlets_dsp)? (i): 0])
#define ftmext_set_connected(e, i, x) ((e)->dsp_connected[i] = (x))

FTMEXT_API void ftmext_init(ftmext_t *ext, ftmext_class_t *cl, fts_patcher_t *scope);
FTMEXT_API void ftmext_delete(ftmext_t *ext);
FTMEXT_API int ftmext_init_get_argument_count(int ac, fts_atom_t *at);

/****************************************************************************
 *
 *  FTM external argument expressions (internal functions)
 *
 */
FTMEXT_API char *ftm_expression_string_from_args(int ac, const fts_atom_t *at);
FTMEXT_API fts_status_t ftmext_expression_parse(ftmext_t *ext, int ac, fts_atom_t *at);
FTMEXT_API fts_status_t ftmext_expression_parse_from_string(ftmext_t *ext, char *str);
FTMEXT_API fts_status_t ftmext_expression_evaluate(ftmext_t *ext, fts_patcher_t *scope, fts_array_t *array, fts_hashtable_t *hash);
FTMEXT_API void ftmext_init_expression(ftmext_t *ext, int ac, fts_atom_t *at);
FTMEXT_API void ftmext_redefine_expression(ftmext_t *ext);

/****************************************************************************
 *
 *  FTM external method implementation
 *
 */

/* union of methods */
typedef union
{
  ftmext_method_number_t n;
  ftmext_method_symbol_t s;
  ftmext_method_object_t o;
  ftmext_method_any_t x;
  ftmext_method_varargs_t v;
} ftmext_method_union_t;

typedef void *ftmext_method_t;

/* internal method macro implementation utilities (used in platform dependent implementations in ftmexternal.h) */
#define _FTMEXT_GEN_METHOD_NAME(p, m) p ## _ ## m
#define _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, t) p ## _ ## m ##  _ ## t

#define _FTMEXT_METHOD_CHECK_SYMBOL(p, m, e, n, a) do { \
  if((n) < 1 || !fts_is_symbol(a)) { ftmext_error(e, "%s: symbol argument required", #m); goto _ftmext_exit_label; } \
} while (0)

#define _FTMEXT_METHOD_CHECK_OBJECT(p, m, e, t, n, a, o) do { \
  fts_class_t *_t = (t); \
  if(n < 1) { ftmext_error(e, "%s: missing FTM object reference", #m); goto _ftmext_exit_label; } \
  if(fts_is_object(a)) { o = fts_get_object(a); } else { ftmext_error(e, "%s: FTM object argument required", #m); goto _ftmext_exit_label; } \
  if(_t != NULL && !fts_object_is_a(o, _t)) { fts_symbol_t cl_name = fts_object_get_class_name(o); \
    if(cl_name != NULL) ftmext_error(e, "%s: not defined with %s argument", #m, fts_symbol_name(cl_name)); \
    else ftmext_error(e, "%s: not defined for given FTM object reference", #m); \
  goto _ftmext_exit_label; } \
} while (0)

#define _FTMEXT_METHOD_CHECK_ANY(p, m, e, n, a) do { \
  if((n) < 1) { ftmext_error(e, "%s: argument required", #m); goto _ftmext_exit_label; } \
} while (0)

/******************************************************************************
 *
 *  FTM external class message declarations
 *
 */
struct _ftmext_message_
{
  fts_symbol_t selector; /**< message selector */
  fts_class_t *type; /**< type ... */
  ftmext_method_union_t meth; /**< meth ... */
  char *doc; /**< doc string */
  struct _ftmext_message_ *next; /**< next message ... */
};

FTMEXT_API void ftmext_class_message(ftmext_class_t *cl, char *name, fts_class_t *type, ftmext_method_t meth, char *doc);

/******************************************************************************
 *
 *  FTM external inlets
 *
 */
struct _ftmext_inlet_
{
  fts_class_t *type; /**< inlet type ... */
  ftmext_method_union_t meth; /**< inlet meth */
  char *doc;
};

FTMEXT_API void ftmext_class_inlet(ftmext_class_t *cl, int i, fts_class_t *type, char *doc);

#define _FTMEXT_INLET_GET_TYPE(e, i) (ftmext_get_class(e)->inlets[(((i) < ftmext_get_class(e)->n_inlets_ctl)? (i): (ftmext_get_class(e)->n_inlets_ctl - 1))].type)
#define _FTMEXT_INLET_ERROR_UNDEFINED(i, e, a) do { ftmext_error((e), "inlet %d: doesn't understand %s", (i), fts_symbol_name(fts_get_class_name(a))); goto _ftmext_exit_label; } while(0)
#define _FTMEXT_INLET_ERROR_WRONG_ARGUMENT(i, e, a, t) do { ftmext_error((e), "inlet %d: doesn't understand %s (%s required)", (i), fts_symbol_name(fts_get_class_name(a)), (t)); goto _ftmext_exit_label; } while(0)
#define _FTMEXT_INLET_ERROR_MISSING_ARGUMENT(i, e) do { ftmext_error((e), "inlet %d: argument missing", (i)); goto _ftmext_exit_label; } while(0)

#define _FTMEXT_INLET_GET_ANY(i, e, n, a) do { if(n < 1) _FTMEXT_INLET_ERROR_MISSING_ARGUMENT(i, e); n = 1; } while(0)
#define _FTMEXT_INLET_GET_NUMBER(i, e, n, a, x) do { _FTMEXT_INLET_GET_ANY(i, e, n, a); if(!fts_is_number(a)) _FTMEXT_INLET_ERROR_WRONG_ARGUMENT(i, e, a, "number"); x = fts_get_number_float(a); } while (0)
#define _FTMEXT_INLET_GET_SYMBOL(i, e, n, a, r) do { _FTMEXT_INLET_GET_ANY(i, e, n, a); if(!fts_is_symbol(a)) _FTMEXT_INLET_ERROR_WRONG_ARGUMENT(i, e, a, "symbol"); r = fts_get_symbol(a); } while (0)
#define _FTMEXT_INLET_GET_OBJECT(i, e, t, n, a, o) do { \
  _FTMEXT_INLET_GET_ANY(i, e, n, a); if(fts_is_object(a)) { o = fts_get_object(a); } else { _FTMEXT_INLET_ERROR_WRONG_ARGUMENT(i, e, a, "FTM object"); } \
  if((t) != NULL && !fts_object_is_a(_o, (t))) _FTMEXT_INLET_ERROR_WRONG_ARGUMENT(i, e, a, fts_symbol_name(fts_class_get_name(t))); \
} while (0)

FTMEXT_API double ftmext_inlet_get_number(ftmext_t *ext, int inlet, int ac, const fts_atom_t *at);
FTMEXT_API fts_object_t *ftmext_inlet_get_object(ftmext_t *ext, int inlet, int ac, const fts_atom_t *at);

/******************************************************************************
 *
 *  FTM external outlets
 *
 */
struct _ftmext_outlet_
{
  fts_class_t *type; /**< outlet type ... */
  char *doc; /**< outlet doc */
};

FTMEXT_API void ftmext_class_outlet(ftmext_class_t *cl, int i, fts_class_t *type, char *doc);

/******************************************************************************
 *
 *  FTM external positional arguments
 *
 */
struct _ftmext_argument_
{
  fts_class_t *type; /**< argument type */
  ftmext_method_union_t meth; /**< meth ...*/
  fts_atom_t def; /**< def ...*/
  char *doc; /**< doc ...*/
};

FTMEXT_API void ftmext_init_arguments(ftmext_t *ext);
FTMEXT_API void ftmext_redefine_arguments(ftmext_t *ext);

FTMEXT_API void ftmext_class_argument(ftmext_class_t *cl, int i, fts_class_t *type, ftmext_method_t meth, fts_atom_t *def, char *doc);

/******************************************************************************
 * 
 *  FTM external attributes
 *
 */
struct _ftmext_attribute_
{
  fts_symbol_t name; /**< attribute name */
  fts_class_t *type; /**< attribute type */
  ftmext_method_union_t meth; /**< attribute method */
  char *doc; /**< attribute doc */
  struct _ftmext_attribute_ *next;
};

FTMEXT_API void ftmext_class_attribute(ftmext_class_t *cl, char *name, fts_class_t *type, ftmext_method_t meth, char *doc);

FTMEXT_API void ftmext_init_attributes(ftmext_t *ext);
FTMEXT_API void ftmext_redefine_attributes(ftmext_t *ext);

#define ftmext_attributes_initialized(e) ((e)->attr_init != 0)

/****************************************************************************
 *
 *  FTM external alarms
 *
 */
FTMEXT_API void ftmext_class_alarm(ftmext_class_t *cl, ftmext_method_void_t callback, char *doc);

/******************************************************************************
 *
 *  FTM external DSP
 *
 */
FTMEXT_API void ftmext_class_dsp(ftmext_class_t *cl);

/******************************************************************************
*******************************************************************************
****
****
****    FTM external public API
****
****
****/

/******************************************************************************
**
**  FTM external class definition public API
**
**/

/** @name Class definition header and basic declarations
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**
* @fn FTMEXT_CLASS(_prefix_ extpackage, _prefix_ extclass, _typedef_struct_ exttype)
 * @brief function header (macro) for external class definition
 * @param extpackage external package prefix (such as ftm, gbr, mnm)
 * @param extclass external class name prefix
 * @param exttype type struct of external (including ftmext_t as first field)
 *
 * This function header is used for the class definition (library main function).
 * The block has to end with FTMEXT_CLASS_RETURN.
 * @code
 
FTMEXT_CLASS(mypkg, myext, myext_t)
{
  < argument declarations >
  < attribute declarations >
  < message declarations >
  < inlet/outlet declarations >
  
  FTMEXT_CLASS_RETURN;
}

 * @endcode
 * @ingroup ftm_externals
 */
FTMEXT_CLASS(_prefix_ extpackage, _prefix_ extclass, _typedef_struct_ exttype);

/**
* @fn FTMEXT_CLASS_NAMED(_prefix_ extpackage, _prefix_ extclass, _typedef_struct_ exttype, char *classname)
 * @brief function header (macro) for external class definition
 * @param extpackage external package prefix (such as ftm, gbr, mnm)
 * @param extclass external class name prefix
 * @param exttype type struct of external (including ftmext_t as first field)
 * @param classname complete external class name (when different from "extpackage.extclass")
 *
 * This function header is used for the class definition (library main function).
 * The block has to be terminated by FTMEXT_CLASS_RETURN.
 * @code
 
FTMEXT_CLASS_NAMED(mypkg, myext, myext_t, "mypkg.example~")
{
  FTMEXT_CLASS_DESCR(_prefix_ extclass, "example class", "This is an example of e class definition.");
  
  < argument declarations >
  < attribute declarations >
  < message declarations >
  < inlet/outlet declarations >
  
  FTMEXT_CLASS_RETURN;
}

 * @endcode
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_NAMED(_prefix_ extpackage, _prefix_ extclass, _typedef_struct_ exttype, char *classname);

/**
* @fn FTMEXT_CLASS_DESCR(_prefix_ extclass, char *breif, char *full)
 * @brief class description (for documentation)
 * @param extclass external class name prefix
 * @param breif breif description (a few words)
 * @param full full description (HTML)
 *
 * This function header is used for DSP class definition.
 * The block has to be terminated by FTMEXT_CLASS_RETURN.
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_DESCR(_prefix_ extclass, char *breif, char *full);

/**
* @fn FTMEXT_CLASS_DSP(_prefix_ extclass)
 * @brief declare class as DSP class
 * @param extclass external class name prefix
 *
 * @code
 
FTMEXT_CLASS_NAMED(mypkg, myext, myext_t, "mypkg.myname")
{
  FTMEXT_CLASS_DESCR(_prefix_ extclass, "example class", "This is an example of e class definition.");
  
  < argument declarations >
  < attribute declarations >
  < message declarations >
  < inlet/outlet declarations >
    
  FTMEXT_CLASS_RETURN;
}

 * @endcode
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_DSP(_prefix_ extclass);

/* additional macros (see platform dependent definitions in ftmexternals.h)
FTMEXT_CLASS_RAW
FTMEXT_CLASS_RETURN;
*/

#endif

/** @} class header and basic declarations */

/** @name Class argument declaration
*  @{ */

#ifdef AVOID_MACROS
/* (see macro definitions below) */

/**
* @fn FTMEXT_CLASS_ARGUMENT_NUMBER(_prefix_ extclass, int idx, _suffix_ meth, float def, char *doc)
 * @brief declare number argument
 * @param extclass external class name prefix
 * @param idx argument index
 * @param meth method name (_suffix_)
 * @param def default value
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ARGUMENT_NUMBER(_prefix_ extclass, int idx, _suffix_ meth, float def, char *doc);

/**
* @fn FTMEXT_CLASS_ARGUMENT_SYMBOL(_prefix_ extclass, int idx, _suffix_ meth, fts_symbol_t def, char *doc)
 * @brief declare symbol argument 
 * @param extclass external class name prefix
 * @param idx argument index
 * @param meth method name (_suffix_)
 * @param def default value
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ARGUMENT_SYMBOL(_prefix_ extclass, int idx, _suffix_ meth, fts_symbol_t def, char *doc);

/**
* @fn FTMEXT_CLASS_ARGUMENT_OBJECT(_prefix_ extclass, fts_class_t *type, int idx, _suffix_ meth, fts_object_t *def, char *doc)
 * @brief declare FTM object argument
 * @param extclass external class name prefix
 * @param type FTM object class (NULL for any)
 * @param idx argument index
 * @param meth method name (_suffix_)
 * @param def default value
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ARGUMENT_OBJECT(_prefix_ extclass, fts_class_t *type, int idx, _suffix_ meth, fts_object_t *def, char *doc);

/**
* @fn FTMEXT_CLASS_ARGUMENT_ANY(_prefix_ extclass, int idx,  _suffix_ meth, fts_atom_t *def, char *doc)
 * @brief declare argument of arbitrary type
 * @param extclass external class name prefix
 * @param idx argument index
 * @param meth method name (_suffix_)
 * @param def default value
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ARGUMENT_ANY(_prefix_ extclass, int idx,  _suffix_ meth, fts_atom_t *def, char *doc);
/* class argument declarations (all arguments at ones) */

/**
* @fn FTMEXT_CLASS_ARGUMENTS_VARARGS(_prefix_ extclass, _suffix_ meth, char *doc)
 * @brief declare argument varargs method to handle all arguments at ones
 * @param extclass external class name prefix
 * @param meth method name (_suffix_)
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ARGUMENTS_VARARGS(_prefix_ extclass, _suffix_ meth, char *doc);

#else

#define FTMEXT_CLASS_ARGUMENT_NUMBER(p, i, m, x, d) do { fts_atom_t _def;  fts_set_float(&_def, (x)); ftmext_class_argument(_c, (i), fts_number_class, _FTMEXT_GEN_METHOD_NAME(p, m), &_def, (d)); } while(0)
#define FTMEXT_CLASS_ARGUMENT_SYMBOL(p, i, m, x, d) do { fts_atom_t _def;  fts_set_symbol(&_def, (x)); ftmext_class_argument(_c, (i), fts_symbol_class, _FTMEXT_GEN_METHOD_NAME(p, m), &_def, (d)); } while(0)
#define FTMEXT_CLASS_ARGUMENT_OBJECT(p, t, i, m, x, d) do { fts_object_t *_x = x; fts_atom_t _def; if(_x != NULL) fts_set_object(&_def, (x)); else fts_set_pointer(&_def, NULL); ftmext_class_argument(_c, (i), (t), _FTMEXT_GEN_METHOD_NAME(p, m), &_def, (d)); } while(0)
#define FTMEXT_CLASS_ARGUMENT_ANY(p, i, m, x, d) do { ftmext_class_argument(_c, (i), ftmext_pseudoclass_any, _FTMEXT_GEN_METHOD_NAME(p, m), (x), (d)); } while(0)
#define FTMEXT_CLASS_ARGUMENTS_VARARGS(p, m, d) do { ftmext_class_argument(_c, 0, ftmext_pseudoclass_varargs, _FTMEXT_GEN_METHOD_NAME(p, m), fts_nix, (d)); } while(0)

#endif

/** @} class argument declarations */

/** @name Class attribute declaration
*  @{ */

#define FTMEXT_ATTRIBUTE_STATIC 0
#define FTMEXT_ATTRIBUTE_DYNAMIC 1

#ifdef AVOID_MACROS
/* (see macro definitions below) */

/**
* @fn FTMEXT_CLASS_ATTRIBUTE_NUMBER(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc)
 * @brief declare number attribute
 * @param extclass external class name prefix
 * @param meth method name (_suffix_)
 * @param dynamic flag (FTMEXT_ATTRIBUTE_STATIC | FTMEXT_ATTRIBUTE_DYNAMIC) whether message is declared with the same name
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ATTRIBUTE_NUMBER(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc);

/**
* @fn FTMEXT_CLASS_ATTRIBUTE_SYMBOL(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc)
 * @brief declare symbol attribute
 * @param extclass external class name prefix
 * @param meth method name (_suffix_)
 * @param dynamic flag (FTMEXT_ATTRIBUTE_STATIC | FTMEXT_ATTRIBUTE_DYNAMIC) whether message is declared with the same name
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ATTRIBUTE_SYMBOL(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc);

/**
* @fn FTMEXT_CLASS_ATTRIBUTE_OBJECT(_prefix_ extclass, fts_class_t *type, _suffix_ meth, int dynamic, char *doc)
 * @brief declare FTM object attribute
 * @param extclass external class name prefix
 * @param type FTM object class (NULL for any)
 * @param meth method name (_suffix_)
 * @param dynamic flag (FTMEXT_ATTRIBUTE_STATIC | FTMEXT_ATTRIBUTE_DYNAMIC) whether message is declared with the same name
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ATTRIBUTE_OBJECT(_prefix_ extclass, fts_class_t *type, _suffix_ meth, int dynamic, char *doc);

/**
* @fn FTMEXT_CLASS_ATTRIBUTE_ANY(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc) 
 * @brief declare attribute of arbitrary value
 * @param extclass external class name prefix
 * @param meth method name (_suffix_)
 * @param dynamic flag (FTMEXT_ATTRIBUTE_STATIC | FTMEXT_ATTRIBUTE_DYNAMIC) whether message is declared with the same name
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ATTRIBUTE_ANY(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc);

/**
* @fn FTMEXT_CLASS_ATTRIBUTE_ANY(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc) 
 * @brief declare varargs attribute
 * @param extclass external class name prefix
 * @param meth method name (_suffix_)
 * @param dynamic flag (FTMEXT_ATTRIBUTE_STATIC | FTMEXT_ATTRIBUTE_DYNAMIC) whether message is declared with the same name
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_ATTRIBUTE_VARARGS(_prefix_ extclass, _suffix_ meth, int dynamic, char *doc);

#else

#define FTMEXT_CLASS_ATTRIBUTE_NUMBER(p, m, x, d) do { ftmext_class_attribute(_c, #m, fts_number_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); if((x) != FTMEXT_ATTRIBUTE_STATIC) FTMEXT_CLASS_MESSAGE_NUMBER(p, m, d); } while(0)
#define FTMEXT_CLASS_ATTRIBUTE_SYMBOL(p, m, x, d) do { ftmext_class_attribute(_c, #m, fts_symbol_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); if((x) != FTMEXT_ATTRIBUTE_STATIC) FTMEXT_CLASS_MESSAGE_SYMBOL(p, m, d); } while(0)
#define FTMEXT_CLASS_ATTRIBUTE_OBJECT(p, t, m, x, d) do { ftmext_class_attribute(_c, #m, t, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); if((x) != FTMEXT_ATTRIBUTE_STATIC) FTMEXT_CLASS_MESSAGE_OBJECT(p, t, m, d); } while(0)
#define FTMEXT_CLASS_ATTRIBUTE_ANY(p, m, x, d) do { ftmext_class_attribute(_c, #m, ftmext_pseudoclass_any, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); if((x) != FTMEXT_ATTRIBUTE_STATIC) FTMEXT_CLASS_MESSAGE_ANY(p, m, d); } while(0)
#define FTMEXT_CLASS_ATTRIBUTE_VARARGS(p, m, x, d) do { ftmext_class_attribute(_c, #m, ftmext_pseudoclass_varargs, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); if((x) != FTMEXT_ATTRIBUTE_STATIC) FTMEXT_CLASS_MESSAGE_VARARGS(p, m, d); } while(0)

#endif

/** @} class attribute declaration */

/** @name Class message declaration
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**    
* @fn FTMEXT_CLASS_MESSAGE_VOID(_prefix_ extclass, _suffix_ meth, char *doc)
* @brief declaration of message without arguments
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_VOID(_prefix_ extclass, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_NUMBER(_prefix_ extclass, _suffix_ meth, char *doc)
* @brief declaration of message with a single number argument
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_NUMBER(_prefix_ extclass, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_SYMBOL(_prefix_ extclass, _suffix_ meth, char *doc)
* @brief declaration of message with a single symbol argument
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_SYMBOL(_prefix_ extclass, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_OBJECT(_prefix_ extclass, fts_class_t *type, _suffix_ meth, char *doc)
* @brief declaration of message with a single FTM object (reference) argument
* @param extclass external class name prefix
* @param type class of object (reference) argument  
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_OBJECT(_prefix_ extclass, fts_class_t *type, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_ANY(_prefix_ extclass, _suffix_ meth, char *doc)
* @brief declaration of message with a single argument of arbitrary type
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_ANY(_prefix_ extclass, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_VARARGS(_prefix_ extclass, _suffix_ meth, char *doc)
* @brief declaration of message with a variable number of arguments
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_VARARGS(_prefix_ extclass, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_ANYTHING(_prefix_ extclass, char *doc)
* @brief declaration of anything message
* @param extclass external class name prefix
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_ANYTHING(_prefix_ extclass, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_BANG(_prefix_ extclass, _suffix_ meth, char *doc)
* @brief declaration of bang message
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_BANG(_prefix_ extclass, _suffix_ meth, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_VOID_NAMED(_prefix_ extclass, _suffix_ meth, char *name, char *doc)
* @brief declaration of message without arguments (with non-canonical name)
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param name message name (selector)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_VOID_NAMED(_prefix_ extclass, _suffix_ meth, char *name, char *doc);

/**    
* @fn FTMEXT_CLASS_MESSAGE_VARARGS_NAMED(_prefix_ extclass, _suffix_ meth, char *name, char *doc)
* @brief declaration of varargs message (with non-canonical name)
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @param name message name (selector)
* @param doc description string
* @ingroup ftm_externals
*/
FTMEXT_CLASS_MESSAGE_VARARGS_NAMED(_prefix_ extclass, _suffix_ meth, char *name, char *doc);

#endif

/** @} class message declarations */

/** @name Class inlet declaration
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**    
* @fn FTMEXT_CLASS_INLET_NUMBER(_prefix_ extclass, int idx, char *doc)
* @brief declaration of an inlet accepting numbers
* @param extclass external class name prefix
* @param idx inlet index
* @param doc description string  
* @ingroup ftm_externals
*/
FTMEXT_CLASS_INLET_NUMBER(_prefix_ extclass, int idx, char *doc);

/**    
* @fn FTMEXT_CLASS_INLET_OBJECT(_prefix_ extclass, fts_class_t *type, int idx, char *doc)
* @brief declaration of an inlet accepting FTM object references
* @param extclass external class name prefix
* @param type FTM object class (NULL for any)
* @param idx inlet index
* @param doc description string  
* @ingroup ftm_externals
*/
FTMEXT_CLASS_INLET_OBJECT(_prefix_ extclass, fts_class_t *type, int idx, char *doc);

/**   
* @fn FTMEXT_CLASS_INLET_ANY(_prefix_ extclass, int idx, char *doc)
* @brief declaration of an inlet accepting values of abrbitrary type
* @param extclass external class name prefix
* @param idx inlet index
* @param doc description string   
* @ingroup ftm_externals
*/
FTMEXT_CLASS_INLET_ANY(_prefix_ extclass, int idx, char *doc);

/**    
* @def FTMEXT_CLASS_INLET_VARARGS(_prefix_ extclass, int idx, char *doc)
* @brief declaration of an inlet accepting lists or single values of arbitrary type
* @param extclass external class name prefix
* @param idx inlet index
* @param doc description string   
* @ingroup ftm_externals
*/
FTMEXT_CLASS_INLET_VARARGS(_prefix_ extclass, int idx, char *doc);

/**    
* @def FTMEXT_CLASS_INLET_MULTI(_prefix_ extclass)
* @brief declaration of a method handling a list corresponding to multiple inlets
* @param extclass external class name prefix
* @param idx inlet index
* @param doc description string   
* @ingroup ftm_externals
*/
FTMEXT_CLASS_INLET_MULTI(_prefix_ extclass);

/**    
* @fn FTMEXT_CLASS_INLET_SIGNAL(_prefix_ extclass, int idx, char *doc)
* @brief declare signal inlet
* @param extclass external class name prefix
* @param idx inlet index
* @param doc description string  
* @ingroup ftm_externals
*/
FTMEXT_CLASS_INLET_SIGNAL(_prefix_ extclass, int idx, char *doc);

#endif

/** @} class inlet declarations */

/** @name Class outlet declaration
*  @{ */

#ifdef AVOID_MACROS
/* (see macro definitions below) */

/**
* @fn FTMEXT_CLASS_OUTLET(_prefix_ extclass, int idx, char *doc)
 * @brief declare arbitrary outlet
 * @param extclass external class name prefix
 * @param idx outlet index
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_OUTLET(_prefix_ extclass, int idx, char *doc);

/**
* @fn FTMEXT_CLASS_OUTLET_BANG(_prefix_ extclass, int idx, char *doc)
 * @brief declare bang outlet
 * @param extclass external class name prefix
 * @param idx outlet index
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_OUTLET_BANG(_prefix_ extclass, int idx, char *doc);

/**
* @fn FTMEXT_CLASS_OUTLET_NUMBER(_prefix_ extclass, int idx, char *doc)
 * @brief declare int outlet 
 * @param extclass external class name prefix
 * @param idx outlet index
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_OUTLET_NUMBER(_prefix_ extclass, int idx, char *doc);

/**
* @fn FTMEXT_CLASS_OUTLET_OBJECT(_prefix_ extclass, fts_class_t *type, int idx, char *doc)
 * @brief declare object outlet
 * @param extclass external class name prefix
 * @param type FTM object class (NULL for any)
 * @param idx outlet index
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_OUTLET_OBJECT(_prefix_ extclass, fts_class_t *type, int idx, char *doc);

/**
* @fn FTMEXT_CLASS_OUTLET_SIGNAL(_prefix_ extclass, int idx, char *doc)
 * @brief declare signal outlet
 * @param extclass external class name prefix
 * @param idx outlet index
 * @param doc description string
 * @ingroup ftm_externals
 */
FTMEXT_CLASS_OUTLET_SIGNAL(_prefix_ extclass, int idx, char *doc);

#else

#define FTMEXT_CLASS_OUTLET(p, i, d) ftmext_class_outlet(_c, (i), NULL, (d))
#define FTMEXT_CLASS_OUTLET_BANG(p, i, d) ftmext_class_outlet(_c, (i), fts_void_class, (d))
#define FTMEXT_CLASS_OUTLET_NUMBER(p, i, d) ftmext_class_outlet(_c, (i), fts_number_class, (d))
#define FTMEXT_CLASS_OUTLET_SYMBOL(p, i, d) ftmext_class_outlet(_c, (i), fts_symbol_class, (d))
#define FTMEXT_CLASS_OUTLET_OBJECT(p, t, i, d) ftmext_class_outlet(_c, (i), (t), (d))
#define FTMEXT_CLASS_OUTLET_ANY(p, i, d) ftmext_class_outlet(_c, (i), ftmext_pseudoclass_any, (d))
#define FTMEXT_CLASS_OUTLET_VARARGS(p, i, d) ftmext_class_outlet(_c, (i), ftmext_pseudoclass_varargs, (d))
#define FTMEXT_CLASS_OUTLET_SIGNAL(p, i, d) ftmext_class_outlet(_c, (i), ftmext_pseudoclass_signal, (d))

/* compatibility */
#define FTMEXT_CLASS_OUTLET_INT(p, i, d) ftmext_class_outlet(_c, (i), fts_number_class, (d))
#define FTMEXT_CLASS_OUTLET_FLOAT(p, i, d) ftmext_class_outlet(_c, (i), fts_number_class, (d))

#endif

/** @} class outlet declarations */

/** @name Class action handler and alarm callback declaration
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**   
* @fn FTMEXT_CLASS_ACTION_OPEN(_prefix_ extclass)
* @brief declare action handler for open (liked to double-click)
* @param extclass external class name prefix
* @ingroup ftm_externals
*/
FTMEXT_CLASS_ACTION_OPEN(_prefix_ extclass);

/**   
* @fn FTMEXT_CLASS_ACTION_SAVE(_prefix_ extclass)
* @brief declare action handler for save
* @param extclass external class name prefix
* @ingroup ftm_externals
*/
FTMEXT_CLASS_ACTION_SAVE(_prefix_ extclass);

/**    
* @fn FTMEXT_CLASS_ALARM(_prefix_ extclass, _suffix_ meth)
* @brief declare alarm with callback (void method)
* @param extclass external class name prefix
* @param meth method name (void method _suffix_)
* @ingroup ftm_externals
*/
FTMEXT_CLASS_ALARM(_prefix_ extclass, _suffix_ meth);

#endif

/** @} class outlet declarations */

/******************************************************************************
**
**  FTM external constructor and deconstructor public API
**
**/

/** @name Constructor and deconstructor (init and delete methods)
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**    
* @fn FTMEXT_INIT(_prefix_ extclass)
* @brief header (marcro) of constructor definition (init method)
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INIT_RETURN.
*/
FTMEXT_INIT(_prefix_ extclass);

/**    
* @fn FTMEXT_DELETE(_prefix_ extclass)
* @brief header (marcro) of de constructor definition (delete method)
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_DELETE_RETURN.
*/
FTMEXT_DELETE(_prefix_ extclass);

/* additional macros (see platform dependent definitions in ftmexternals.h)
FTMEXT_INIT_RETURN
FTMEXT_DELETE_RETURN
*/

#endif

/**
* @fn void ftmext_set_inlets(ftmext_t *ext, int n)
 * @brief set number of signal inlets in constructor (init method) 
 * @param ext pointer to external object (self)
 * @param n number of inlets
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_set_inlets(ftmext_t *obj, int n);

/**
* @fn void ftmext_set_outlets(ftmext_t *ext, int n)
 * @brief set number of outlets in constructor (init method) 
 * @param ext pointer to external object (self)
 * @param n number of outlets
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_set_outlets(ftmext_t *ext, int n);

/**
* @fn void ftmext_set_dsp_outlets(ftmext_t *ext, int n)
 * @brief set number of signal outlets in constructor (init method) 
 * @param ext pointer to external object (self)
 * @param n number of signal outlets
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_set_dsp_outlets(ftmext_t *ext, int n);

/** @}	end of doc group constructor and deconstructor */

/******************************************************************************
**
**  FTM external method definition public API
**
**/

/** @name Method definition (for arguments, attributes and messages)
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**    
* @fn FTMEXT_METHOD_VOID(_prefix_ extclass, _suffix_ meth)
* @brief header (macro) of argumentless method definition
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_VOID(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  
  < method implementation >
    
    FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_VOID(_prefix_ extclass, _suffix_ meth);

/**    
* @fn FTMEXT_METHOD_NUMBER(_prefix_ extclass, _suffix_ meth)
* @brief header (macro) of number method definition
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_NUMBER(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  int intarg = FTMEXT_GET_INT();
  double intarg = FTMEXT_GET_FLOAT();
  
  < method implementation >
    
  FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_NUMBER(_prefix_ extclass, _suffix_ meth);

/**    
* @fn FTMEXT_METHOD_SYMBOL(_prefix_ extclass, _suffix_ meth)
* @brief header (macro) of symbol method definition
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_SYMBOL(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  fts_symbol_t symarg = FTMEXT_GET_SYMBOL();
  
  < method implementation >
    
  FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_SYMBOL(_prefix_ extclass, _suffix_ meth);

/**    
* @fn FTMEXT_METHOD_OBJECT(_prefix_ extclass, fts_class_t *type, _suffix_ meth)
* @brief header (macro) of object method definition
* @param extclass external class name prefix
* @param type FTM object class 
* @param meth method name (_suffix_)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_OBJECT(myext, type)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  fts_symbol_t objarg = FTMEXT_GET_OBJECT();
  
  < method implementation >
    
  FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_OBJECT(_prefix_ extclass, fts_class_t *type, _suffix_ meth);

/**    
* @fn FTMEXT_METHOD_ANY(_prefix_ extclass, _suffix_ meth)
* @brief header (macro) of method definition with single argument of abitrary type
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_ANY(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  fts_atom_t *atomarg = FTMEXT_GET_ANY();
  
  < method implementation >
    
  FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_ANY(_prefix_ extclass, _suffix_ meth);

/**    
* @fn FTMEXT_METHOD_VARARGS(_prefix_ extclass, _suffix_ meth)
* @brief header (macro) of varargs method definition
* @param extclass external class name prefix
* @param meth method name (_suffix_)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_VARARGS(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  
  < method implementation >
    
  FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_VARARGS(_prefix_ extclass, _suffix_ meth);

/**    
* @fn FTMEXT_METHOD_ANYTHING(_prefix_ extclass)
* @brief header (macro) of anything method definition
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_METHOD_RETURN.
* @code

FTMEXT_METHOD_ANYTHING(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  fts_symbol_t s = FTMEXT_GET_SELECTOR();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  
  < method implementation >
    
  FTMEXT_METHOD_RETURN;
}

* @endcode
*/
FTMEXT_METHOD_ANYTHING(_prefix_ extclass);      

/* additional macros (see platform dependent definitions in ftmexternals.h)
FTMEXT_METHOD_RETURN;
FTMEXT_EXIT; */

#endif

/** @}	end of doc group method definition */

/******************************************************************************
**
**  FTM external inlet method definition public API
**
**/

/** @name Inlet method definition
*  @{ */

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h)*/

/**    
* @fn FTMEXT_INLET_NUMBER(_prefix_ extclass)
* @brief header (macro) of method definition for number inlets
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INLET_RETURN.
* @code

FTMEXT_INLET_NUMBER(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  int intval = FTMEXT_GET_INT();
  double intval = FTMEXT_GET_FLOAT();
  
  < method implementation >
    
  FTMEXT_INLET_RETURN;
}

* @endcode
*/
FTMEXT_INLET_NUMBER(_prefix_ extclass);

/**
* @fn void FTMEXT_INLET_NUMBER_ERROR()
 * @brief inlet error message (used in number inlet definition)
 * @ingroup ftm_externals
 */
FTMEXT_INLET_NUMBER_ERROR();

/**    
* @fn FTMEXT_INLET_OBJECT(_prefix_ extclass, fts_class_t *type)
* @brief header (macro) of method definition for FTM object inlets
* @param extclass external class name prefix
* @param type FTM object class (NULL for any)
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INLET_RETURN.
* @code

FTMEXT_METHOD_OBJECT(myext, type)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  fts_symbol_t objarg = FTMEXT_GET_OBJECT();
  
  < method implementation >
    
  FTMEXT_INLET_RETURN;
}

* @endcode
*/
FTMEXT_INLET_OBJECT(_prefix_ extclass, fts_class_t *type);

/**
* @fn FTMEXT_INLET_OBJECT_ERROR()
 * @brief inlet error message (used in object inlet definition)
 * @ingroup ftm_externals
 */
FTMEXT_INLET_OBJECT_ERROR();

/**    
* @fn FTMEXT_INLET_ANY(_prefix_ extclass)
* @brief header (macro) of method definition for arbitrary value inlets
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INLET_RETURN.
* @code

FTMEXT_METHOD_ANY(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  fts_atom_t *atomarg = FTMEXT_GET_ANY();
  
  < method implementation >
    
  FTMEXT_INLET_RETURN;
}

* @endcode
*/  
FTMEXT_INLET_ANY(_prefix_ extclass);

/**
* @fn FTMEXT_INLET_ANY_ERROR()
 * @brief inlet error message (used in any inlet definition)
 * @ingroup ftm_externals
 */
FTMEXT_INLET_ANY_ERROR();

/**    
* @fn FTMEXT_INLET_VARARGS(_prefix_ extclass)
* @brief header (macro) of method definition for varargs (list) inlets
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INLET_RETURN.
* @code

FTMEXT_INLET_VARARGS(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  
  < method implementation >
    
  FTMEXT_INLET_RETURN;
}

* @endcode
*/  
FTMEXT_INLET_VARARGS(_prefix_ extclass);

/**    
* @fn FTMEXT_INLET_VARARGS_CLASSIC(_prefix_ extclass)
* @brief header (macro) of method definition for varargs (list) inlets
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INLET_RETURN.
* @code

FTMEXT_INLET_VARARGS_CLASSIC(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  
  < method implementation >
    
  FTMEXT_INLET_RETURN;
}

* @endcode
*/  
FTMEXT_INLET_VARARGS_CLASSIC(_prefix_ extclass);

/**
* @fn FTMEXT_INLET_VARARGS_ERROR()
 * @brief inlet error message (used in varargs inlet definition)
 * @ingroup ftm_externals
 */ 
FTMEXT_INLET_VARARGS_ERROR();

/**    
* @fn FTMEXT_INLET_MULTI(_prefix_ extclass)
* @brief header (macro) of method definition for inlets handling a list corresponding to multiple inlets
* @param extclass external class name prefix
* @ingroup ftm_externals
*
* The block has to be terminated by FTMEXT_INLET_RETURN.
* @code

FTMEXT_INLET_MULTI(myext)
{
  ftmext_t *self = FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  
  < method implementation >
    
  FTMEXT_INLET_RETURN;
}

* @endcode
*/  
FTMEXT_INLET_MULTI(_prefix_ extclass);

/* additional macros (see platform dependent definitions in ftmexternals.h)
FTMEXT_INLET_RETURN;
*/

#else

#define FTMEXT_INLET_NUMBER_ERROR() do { ftmext_error(_e, "inlet %d: doesn't understand int or float", _wi); } while(0)
#define FTMEXT_INLET_OBJECT_ERROR()  do { \
  if(_t == NULL) ftmext_error(_e, "inlet %d doesn't understand FTM object reference", _wi); \
    else ftmext_error(_e, "inlet %d: doesn't understand FTM %s reference", _wi, fts_symbol_name(fts_class_get_name(_t))); \
} while(0)
#define FTMEXT_INLET_ANY_ERROR() do { ftmext_error(_e, "inlet %d: invalid value", _wi); } while(0)
#define FTMEXT_INLET_VARARGS_ERROR() do { ftmext_error(_e, "inlet %d: doesn't understand varargs", _wi); } while(0)

#endif

/** @}	end of doc group inlet method definition */

/******************************************************************************
**
**  FTM external handler method definition public API
**
**/

/** @name Handler method definition
*  @{ */

FTMEXT_API void ftmext_class_save_handler(ftmext_class_t *cl, ftmext_save_handler_t handler);
FTMEXT_API void ftmext_save(ftmext_t *ext, void *handle);

FTMEXT_API void ftmext_save_message(ftmext_t *ext, fts_symbol_t message, int ac, const fts_atom_t *at);

#ifdef AVOID_MACROS
/* (see platform dependent definitions in ftmexternals.h) */

/**   
* @fn FTMEXT_HANDLER_SAVE(_prefix_ extclass)
* @brief header of save action handler definiton
* @param extclass external class name prefix.
* @ingroup ftm_externals
*/
FTMEXT_HANDLER_SAVE(_prefix_ extclass);

/**   
* @fn FTMEXT_HANDLER_OPEN(_prefix_ extclass)
* @brief header of open action handler definiton
* @param extclass external class name prefix
* @ingroup ftm_externals
*/
FTMEXT_HANDLER_OPEN(_prefix_ extclass);

/* additional macros (see platform dependent definitions in ftmexternals.h)
FTMEXT_HANDLER_RETURN;
*/

#endif

/** @}	end of doc group handler method definition */

/******************************************************************************
**
**  FTM external method utlilities public API
**
**/

/** @name Method utlilities
*  @{ */

#ifdef AVOID_MACROS
/* (see macro definitions below) */

/**
* @fn ftmext_t *FTMEXT_GET_EXT()
 * @brief get pointer to ftmext object (self)
 * @return pointer to ftmext
 * @ingroup ftm_externals
 */
ftmext_t *FTMEXT_GET_EXT();

/**
* @fn int FTMEXT_GET_INT()
 * @brief get int argument (in number methods)
 * @return int argument
 * @ingroup ftm_externals
 */
int FTMEXT_GET_INT();

/**
* @fn float FTMEXT_GET_FLOAT()
 * @brief get float argument (in number methods)
 * @return float argument
 * @ingroup ftm_externals
 */
double FTMEXT_GET_FLOAT();

/**
* @fn float FTMEXT_GET_NUMBER()
 * @brief get number (float) argument (in number methods)
 * @return number (float) argument
 * @ingroup ftm_externals
 */
double FTMEXT_GET_NUMBER();

/**
* @fn fts_symbol_t FTMEXT_GET_SYMBOL()
 * @brief get symbol argument (in symbol methods)
 * @return symbol argument
 * @ingroup ftm_externals
 */
fts_symbol_t FTMEXT_GET_SYMBOL();

/**
* @fn fts_object_t *FTMEXT_GET_OBJECT()
 * @brief get fts_object argument (in object methods)
 * @return fts_object argument 
 * @ingroup ftm_externals
 */
fts_object_t *FTMEXT_GET_OBJECT();

/**
* @fn fts_atom_t *FTMEXT_GET_ANY()
 * @brief get single atom argument (in any methods)
 * @return atom argument
 * @ingroup ftm_externals
 */
fts_atom_t *FTMEXT_GET_ANY();

/**
* @fn int FTMEXT_GET_COUNT()
 * @brief get argument count (in varargs and anything methods and external init)
 * @return argument count
 * @ingroup ftm_externals
 */
int FTMEXT_GET_COUNT();

/**
* @fn fts_atom_t *FTMEXT_GET_ARGS()
 * @brief get multiple arguments array (in varargs and anything methods and external init)
 * @return pointer to arguments array
 * @ingroup ftm_externals
 */
fts_atom_t *FTMEXT_GET_ARGS();

/**
* @fn fts_symbol_t FTMEXT_GET_SELECTOR()
 * @brief get method selector (in anything methods)
 * @return message selector
 * @ingroup ftm_externals
 */
fts_symbol_t FTMEXT_GET_SELECTOR();

/**
* @fn int FTMEXT_GET_INLET()
 * @brief get inlet index (in inlet methods)
 * @return index of current inlet
 * @ingroup ftm_externals
 */
int FTMEXT_GET_INLET();

#else

#define FTMEXT_GET_EXT() (_e)
#define FTMEXT_GET_INT() ((int)_x)
#define FTMEXT_GET_FLOAT() (_x)
#define FTMEXT_GET_NUMBER() (_x)
#define FTMEXT_GET_SYMBOL() (_r)
#define FTMEXT_GET_OBJECT() (_o)
#define FTMEXT_GET_ANY() (_a)
#define FTMEXT_GET_COUNT() (_n)
#define FTMEXT_GET_ARGS() (_a)
#define FTMEXT_GET_SELECTOR() (_s)
#define FTMEXT_GET_INLET() (_wi)

#endif

/**
* @fn void ftmext_outlet_bang(ftmext_t *ext, int w)
 * @brief send bang from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_bang(ftmext_t *ext, int w);

/**
* @fn void ftmext_outlet_int(ftmext_t *ext, int w, int i)
 * @brief send int from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param i int value
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_int(ftmext_t *ext, int w, int i);

/**
* @fn void ftmext_outlet_float(ftmext_t *ext, int w, double f)
 * @brief send float from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param f float value
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_float(ftmext_t *ext, int w, double f);

/**
* @fn void ftmext_outlet_number(ftmext_t *ext, int w, double f)
 * @brief send number (float) from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param f number (float) value
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_number(ftmext_t *ext, int w, double f);

/**
* @fn void ftmext_outlet_symbol(ftmext_t *ext, int w, fts_symbol_t s)
 * @brief send symbol from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param s symbol value
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_symbol(ftmext_t *ext, int w, fts_symbol_t s);

/**
* @fn void ftmext_outlet_object(ftmext_t *ext, int w, fts_object_t *o)
 * @brief send FTM object reference from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param o reference to FTM object (fts_object_t *)
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_object(ftmext_t *ext, int w, fts_object_t *o);

/**
* @fn void ftmext_outlet_any(ftmext_t *ext, int w, fts_atom_t *a)
 * @brief send any value from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param a arbitrary values
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_any(ftmext_t *ext, int w, fts_atom_t *a);

/**
* @fn void ftmext_outlet_varargs(ftmext_t *ext, int w, int n, fts_atom_t *a)
 * @brief send varargs (list) from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param n number of values
 * @param a array of values
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_varargs(ftmext_t *ext, int w, int n, fts_atom_t *a);

/**
* @fn void ftmext_outlet_message(ftmext_t *ext, int w, fts_symbol_t s, int n, fts_atom_t *a)
 * @brief send message from given outlet
 * @param ext pointer to external object (self)
 * @param w outlet index (leftmost inlet is 0)
 * @param s message selector
 * @param n number of message arguments
 * @param a array of message arguments
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_outlet_message(ftmext_t *ext, int w, fts_symbol_t s, int n, fts_atom_t *a);

/**
* @fn ftmext_alarm_set(ftmext_t *ext, double delay)
 * @brief set alarm to given delay
 * @param ext pointer to external object (self)
 * @param delay alarm delay time
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_alarm_set(ftmext_t *ext, double delay);

/**
* @fn ftmext_alarm_reset(ftmext_t *ext)
 * @brief reset alarm (set to inifinit delay time)
 * @param ext pointer to external object (self)
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_alarm_reset(ftmext_t *ext);

/**
* @fn void ftmext_post(ftmext_t *ext, const char *format, ...)
 * @brief posts error message to the console
 * @param ext pointer to external object (self)
 * @param format message (printf format)
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_post(ftmext_t *obj, const char *format, ...);

/**
* @fn void ftmext_error(ftmext_t *ext, const char *format, ...)
 * @brief posts message to the console
 * @param ext pointer to external object (self)
 * @param format error message (printf format)
 * @ingroup ftm_externals
 */
FTMEXT_API void ftmext_error(ftmext_t *obj, const char *format, ...);

/** @}	end of doc group method utlilities */

/******************************************************************************
**
**  FTM external DSP method definition public API
**
**/

/** @name DSP method definition
*  @{ */

#ifdef AVOID_MACROS
/* DSP related methods and functions (see platform dependent definitions in ftmexternals.h) */

/**    
* @fn FTMEXT_DSP_INIT(_prefix_ extclass)
* @brief header (macro) of DSP init method definition (method called at DSP start)
* @param extclass external class name prefix
* @ingroup ftm_externals
*/
FTMEXT_DSP_INIT(_prefix_ extclass);

/**    
* @fn int FTMEXT_GET_DSP_SR()
* @brief get DSP sample rate (within DSP init method)
* @return DSP sample rate
* @ingroup ftm_externals
*/
int FTMEXT_GET_DSP_SR();

/**    
* @fn int FTMEXT_GET_DSP_SIZE()
* @brief get DSP vector size (within DSP init method)
* @return DSP vector size
* @ingroup ftm_externals
*/
int FTMEXT_GET_DSP_SIZE(); 

/**   
* @fn FTMEXT_DSP_PERFORM(_prefix_ extclass)
* @brief header (macro) of DSP perform method definition
* @param extclass external class name prefix
* @ingroup ftm_externals
*/
FTMEXT_DSP_PERFORM(_prefix_ extclass);

/**   
* @fn float *FTMEXT_GET_DSP_INPUT(int idx)
* @brief get input signal vector (within DSP perform method)
* @param idx DSP inlet index
* @return pointer to signal vector
* @ingroup ftm_externals
*/
float *FTMEXT_GET_DSP_INPUT(int idx);

/**   
* @fn float *FTMEXT_GET_DSP_OUTPUT(int idx)
* @brief get output signal vector (within DSP perform method)
* @param idx DSP outlet index
* @return pointer to signal vector
* @ingroup ftm_externals
*/
float *FTMEXT_GET_DSP_OUTPUT(int idx);

/**   
* @fn FTMEXT_DSP_INLET_CONNECTED(int idx)
* @brief test whether a given inlet is connected to a signal connection
* @param idx DSP inlet index
* @return boolean connected
* @ingroup ftm_externals
*/
int FTMEXT_DSP_INLET_CONNECTED(int idx);

/* additional macros (see platform dependent definitions in ftmexternals.h)
FTMEXT_DSP_INIT_RETURN
FTMEXT_DSP_PERFORM_RETURN
*/
#else

#define FTMEXT_DSP_INLET_CONNECTED(i) (ftmext_get_connected(_e, i))

#endif

/** @}	end of doc group DSP method definition */

/**************************************************************************************
 *
 *  FTM run-time environment dependent stuff
 *
 */
#include "ftmexternals.h"

#endif
