/*
 * FTM
 * Copyright (C) 2004 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

#ifndef _FTMEXTERNALS_H
#define _FTMEXTERNALS_H

#include "fts.h"

/****************************************************************************
 *
 *  FTM Max/MSP external class definition
 *
 */

/* Max/MSP specific functions (defined in maxext.c) */
FTMEXT_API void maxext_setup_failed(void);
FTMEXT_API void maxext_restore_ftm(ftmext_t *ext, fts_symbol_t s, short ac, fts_atom_t *at);
FTMEXT_API void maxext_restore_scope(ftmext_t *ext, fts_symbol_t s, short ac, fts_atom_t *at);
FTMEXT_API void maxext_restore_absargs(ftmext_t *ext, fts_symbol_t s, short ac, fts_atom_t *at);
FTMEXT_API void maxext_save(ftmext_t *ext, void *binbuf);
FTMEXT_API void maxext_assist(ftmext_t *self, void *box, long msg, long arg, char *dst);
FTMEXT_API void maxext_redefine(ftmext_t *ext);

#ifndef AVOID_MACROS

#define FTMEXT_CLASS_RAW(pp, p, t, n) int main (void) { ftmext_class_t *_c = NULL; \
if (!ftmlib_init(FTM_VERSION_MAJOR, FTM_VERSION_MINOR)) { maxext_setup_failed();  return 0; } \
  _c = ftmext_class_new(fts_new_symbol(n)); \
  setup((t_messlist **) &(((_c))->handle), (method) (_FTMEXT_GEN_METHOD_NAME(p, init)), (method) (_FTMEXT_GEN_METHOD_NAME(p, delete)), (short) sizeof(t), 0, A_GIMME, 0); \
  addmess((method)maxext_restore_ftm, ftm_str_restore_ftm, A_GIMME, 0); \
  addmess((method)maxext_restore_scope, ftm_str_restore_scope, A_GIMME, 0); \
  addmess((method)maxext_restore_absargs, ftm_str_restore_absargs, A_GIMME, 0); \
  addmess((method)maxext_save, "save", A_CANT, 0);  
  
#define FTMEXT_CLASS(pp, p, t) _FTMEXT_CLASS(pp, p, t, #pp "." #p)
#define FTMEXT_CLASS_NAMED(pp, p, t, n) _FTMEXT_CLASS(pp, p, t, n)
#define FTMEXT_CLASS_DSP(p) do { dsp_initclass(); ftmext_class_dsp(_c); addmess((method)_FTMEXT_GEN_METHOD_NAME(p, _dsp), "dsp", A_CANT, 0); } while(0)
#define FTMEXT_CLASS_DESCR(p, b, f) do { ftmext_class_descr(_c, (b), (f)); } while(0)
#define FTMEXT_CLASS_RETURN _ftmext_exit_label: return 1; }

#endif

#define _FTMEXT_CLASS(pp, p, t, n) FTMEXT_CLASS_RAW(pp, p, t, n) \
  addmess((method)maxext_redefine, ftm_str_redefine, A_CANT, 0); \
  addmess((method)maxext_assist, "assist", A_CANT, 0);


/****************************************************************************
 *
 *  FTM Max/MSP external external constructor/deconstructor definition
 *
 */
/* Max/MSP specific functions (defined in maxext.c) */
FTMEXT_API ftmext_t *maxext_init(fts_symbol_t name, t_patcher *patcher, short *ac, fts_atom_t **at);
FTMEXT_API int maxext_init_raw(ftmext_t *ext, ftmext_class_t *cl, t_patcher *patcher, short *rac, fts_atom_t **rat);
FTMEXT_API void maxext_init_finalize(ftmext_t *ext, short ac, fts_atom_t *at);
FTMEXT_API void maxext_delete(ftmext_t *ext);

#ifndef AVOID_MACROS

#define FTMEXT_INIT_RAW(p) static void *_FTMEXT_GEN_METHOD_NAME(p, init) (fts_symbol_t _s, short _nn, fts_atom_t *_a) { \
  ftmext_class_t *_c = ftmext_class_get_by_name(_s); ftmext_t *_e = newobject(_c->handle); \
  int _n = maxext_init_raw(_e, _c, (t_patcher *)(gensym("#P")->s_thing), &_nn, &_a); \
  if(_e == NULL || _n < 0) return NULL; /* "_n >= 0" fake condition to avoid compiler warning "unused variable" */

#define FTMEXT_INIT(p) static void *_FTMEXT_GEN_METHOD_NAME(p, init) (fts_symbol_t _s, short _nn, fts_atom_t *_a) { \
  ftmext_t *_e = maxext_init(_s, (t_patcher *)(gensym("#P")->s_thing), &_nn, &_a); \
  int _n = ftmext_init_get_argument_count(_nn, _a); \
  if(_e == NULL || _n < 0) return NULL; /* "_n >= 0" fake condition to avoid compiler warning "unused variable" */
  
#define FTMEXT_INIT_RETURN \
  maxext_init_finalize(_e, (int)_nn, _a); \
  _ftmext_exit_label: return(_e); }

#define FTMEXT_DELETE(p) static void _FTMEXT_GEN_METHOD_NAME(p, delete) (ftmext_t *_e) { \
  maxext_delete(_e);
    
#define FTMEXT_DELETE_RETURN _ftmext_exit_label: ; }

#endif

/*********************************************************************************
 *
 *  FTM Max/MSP external method definitions
 *
 */
#ifndef AVOID_MACROS

#define FTMEXT_METHOD_VOID(p, m) static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e) {
#define FTMEXT_METHOD_NUMBER(p, m) static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, double _x) {
#define FTMEXT_METHOD_SYMBOL(p, m) static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, fts_symbol_t _r); \
  static void _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, symbol) (ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) { \
    _FTMEXT_METHOD_CHECK_SYMBOL(p, m, _e, _n, _a); _FTMEXT_GEN_METHOD_NAME(p, m) (_e, fts_get_symbol(_a)); _ftmext_exit_label: ; } \
  static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, fts_symbol_t _r) {
#define FTMEXT_METHOD_OBJECT(p, t, m) static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, fts_object_t *_o); \
  static void _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, object) (ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) { fts_object_t *_o = NULL; \
    _FTMEXT_METHOD_CHECK_OBJECT(p, m, _e, t, _n, _a, _o); _FTMEXT_GEN_METHOD_NAME(p, m) (_e, _o); _ftmext_exit_label: ; } \
  static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, fts_object_t *_o) {
#define FTMEXT_METHOD_ANY(p, m) static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, fts_atom_t *_a); \
  static void _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, any) (ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) { \
    _FTMEXT_METHOD_CHECK_ANY(p, m, _e, _n, _a); _FTMEXT_GEN_METHOD_NAME(p, m) (_e, _a); _ftmext_exit_label: ; } \
  static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, fts_atom_t *_a) {
#define FTMEXT_METHOD_VARARGS(p, m) static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, int _n, fts_atom_t *_a); \
  static void _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, varargs) (ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) { _FTMEXT_GEN_METHOD_NAME(p, m)(_e, _n, _a); } \
  static void _FTMEXT_GEN_METHOD_NAME(p, m) (ftmext_t *_e, int _n, fts_atom_t *_a) {
#define FTMEXT_METHOD_ANYTHING(p) static void _FTMEXT_GEN_METHOD_NAME(p, anything) (ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) {        
#define FTMEXT_METHOD_RETURN _ftmext_exit_label: ; }

/* generic return from all ftmext_ methods and functions */
#define FTMEXT_EXIT goto _ftmext_exit_label

#endif

/****************************************************************************
 *
 *  FTM Max/MSP external class message declarations
 *
 */
#ifndef AVOID_MACROS
   
#define FTMEXT_CLASS_MESSAGE_VOID(p, m, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME(p, m), #m, 0); ftmext_class_message(_c, #m, fts_void_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_NUMBER(p, m, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME(p, m), #m, A_FLOAT, 0); ftmext_class_message(_c, #m, fts_number_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_SYMBOL(p, m, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, symbol), #m, A_GIMME, 0); ftmext_class_message(_c, #m, fts_symbol_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_OBJECT(p, t, m, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, object),  #m, A_GIMME, 0); ftmext_class_message(_c, #m, (t), _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_ANY(p, m, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, any), #m, A_GIMME, 0); ftmext_class_message(_c, #m, ftmext_pseudoclass_any, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_VARARGS(p, m, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, varargs), #m, A_GIMME, 0); ftmext_class_message(_c, #m, ftmext_pseudoclass_varargs, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_ANYTHING(p, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME(p, anything), "anything", A_GIMME, 0); ftmext_class_message(_c, "", ftmext_pseudoclass_anything, _FTMEXT_GEN_METHOD_NAME(p, anything), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_VOID_NAMED(p, m, n, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME(p, m), n, 0); ftmext_class_message(_c, #m, fts_void_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_VARARGS_NAMED(p, m, n, d) do { addmess((method) _FTMEXT_GEN_METHOD_NAME_TYPE(p, m, varargs), n, A_GIMME, 0); ftmext_class_message(_c, #m, ftmext_pseudoclass_varargs, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
#define FTMEXT_CLASS_MESSAGE_BANG(p, m, d) do { addbang((method) _FTMEXT_GEN_METHOD_NAME(p, m)); ftmext_class_message(_c, "bang", fts_void_class, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)

#endif

/****************************************************************************
 *
 *  FTM Max/MSP external inlets
 *
 */
#ifndef AVOID_MACROS
    
#define _FTMEXT_CLASS_INLET(p, t, i, d) do { \
  addint((method) _ ## p ## _inlet_int); \
  addfloat((method) _ ## p ## _inlet_float); \
  /* addmess((method) _ ## p ## _inlet_bang, "bang", 0); */ \
  addmess((method) _ ## p ## _inlet_symbol, "symbol", A_SYM, 0); \
  addmess((method) _ ## p ## _inlet_object, ftm_str_obj, A_GIMME, 0); \
  addmess((method) _ ## p ## _inlet_list, "list", A_GIMME, 0); \
  ftmext_class_inlet(_c, (i), (t), (d)); \
} while(0)

#define _MAXEXT_GET_INLET(e) (proxy_getinlet((t_object *)(e)))
    
/* class inlet declarations */
#define FTMEXT_CLASS_INLET_BANG(p, i, d) _FTMEXT_CLASS_INLET(p, fts_void_class, (i), (d))
#define FTMEXT_CLASS_INLET_NUMBER(p, i, d) _FTMEXT_CLASS_INLET(p, fts_number_class, (i), (d))
#define FTMEXT_CLASS_INLET_SYMBOL(p, i, d) _FTMEXT_CLASS_INLET(p, fts_symbol_class, (i), (d))
#define FTMEXT_CLASS_INLET_OBJECT(p, t, i, d) _FTMEXT_CLASS_INLET(p, t, (i), (d))
#define FTMEXT_CLASS_INLET_ANY(p, i, d) _FTMEXT_CLASS_INLET(p, ftmext_pseudoclass_any, (i), (d))
#define FTMEXT_CLASS_INLET_VARARGS(p, i, d) _FTMEXT_CLASS_INLET(p, ftmext_pseudoclass_varargs, (i), (d))
#define FTMEXT_CLASS_INLET_MULTI(p) _FTMEXT_CLASS_INLET(p, ftmext_pseudoclass_varargs, -1, NULL)
#define FTMEXT_CLASS_INLET_SIGNAL(p, i, d) ftmext_class_inlet(_c, (i), ftmext_pseudoclass_signal, (d))

#define FTMEXT_INLET(p) static void p ## _inlet(ftmext_t *_e, int _wi, int _n, fts_atom_t *_a); \
  /* static void _ ## p ## _inlet_bang(ftmext_t *_e) { p ## _inlet(_e, _MAXEXT_GET_INLET(_e), 0, NULL); } */ \
  static void _ ## p ## _inlet_int(ftmext_t *_e, int _i) { fts_atom_t _a; fts_set_int(&_a, _i); p ## _inlet(_e, _MAXEXT_GET_INLET(_e), 1, &_a); } \
  static void _ ## p ## _inlet_float(ftmext_t *_e, double _f) { fts_atom_t _a; fts_set_float(&_a, _f); p ## _inlet(_e, _MAXEXT_GET_INLET(_e), 1, &_a); } \
  static void _ ## p ## _inlet_symbol(ftmext_t *_e, fts_symbol_t _r) { fts_atom_t _a; fts_set_symbol(&_a, _r); p ## _inlet(_e, _MAXEXT_GET_INLET(_e), 1, &_a); } \
  static void _ ## p ## _inlet_object(ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) { int _wi = _MAXEXT_GET_INLET(_e); \
    if(fts_is_number(_a)) { int _id = fts_get_number_int(_a); fts_object_t *_o = fts_client_get_object(_id); fts_atom_t _b; \
	  if(_o == NULL) { ftmext_error(_e, "inlet %d: invalid FTM object id: %d", _wi, _o); } fts_set_object(&_b, _o);	p ## _inlet(_e, _wi, 1, &_b); } \
	else { p ## _inlet(_e, _wi, 1, _a); } } \
  static void _ ## p ## _inlet_list(ftmext_t *_e, fts_symbol_t _s, int _n, fts_atom_t *_a) { \
	if(!_e->cl->inlet_varargs_dispatch) { p ## _inlet(_e, _MAXEXT_GET_INLET(_e), _n, _a); } \
	else { int _wi = _MAXEXT_GET_INLET(_e); int _i; if(_wi + _n > _e->n_inlets_ctl) _n = _e->n_inlets_ctl - _wi; for(_i=_n-1; _i>=0; _i--) p ## _inlet(_e, _wi + _i, 1, _a + _i); } } \
  static void p ## _inlet(ftmext_t *_e, int _wi, int _n, fts_atom_t *_a) { \
    double _x = 0.0; fts_symbol_t _r = NULL; fts_object_t *_o = NULL; \
    fts_class_t *_t = _FTMEXT_INLET_GET_TYPE(_e, _wi); int _tid = 99; if(_t != NULL) _tid = fts_class_get_id(_t); \
    switch (_tid) { \
      case 0: _FTMEXT_INLET_ERROR_UNDEFINED(_wi, _e, _a); break; \
      case FTS_TYPEID_VOID: break; \
      case FTS_TYPEID_NUMBER: _FTMEXT_INLET_GET_NUMBER(_wi, _e, _n, _a, _x); break; \
      case FTS_TYPEID_SYMBOL: _FTMEXT_INLET_GET_SYMBOL(_wi, _e, _n, _a, _r); break; \
      case FTS_TYPEID_STRING: goto _ftmext_exit_label; \
      case FTS_TYPEID_POINTER: goto _ftmext_exit_label; \
      case FTMEXT_TYPEID_ANY: _FTMEXT_INLET_GET_ANY(_wi, _e, _n, _a); break; \
      case FTMEXT_TYPEID_VARARGS: break; \
      default: _FTMEXT_INLET_GET_OBJECT(_wi, _e, _t, _n, _a, _o); break; \
    }
    
#define FTMEXT_INLET_RETURN _ftmext_exit_label: ; }
    
#endif

/****************************************************************************
 *
 *  FTM Max/MSP external DSP method definition
 *
 */
/* Max/MSP specific functions (defined in maxext.c) */
FTMEXT_API void maxext_dsp_add(ftmext_t *ext, void *perform, t_signal **sp, int n);

#ifndef AVOID_MACROS

/* dsp initialization */
#define FTMEXT_DSP_INIT(p) static void _FTMEXT_GEN_METHOD_NAME(p, _dsp) (ftmext_t *_e, t_signal **_sp, short *count) { \
  int _dsp_size = _sp[0]->s_n; double _dsp_sr = _sp[0]->s_sr; void *_pr = _FTMEXT_GEN_METHOD_NAME(p, _perform); int _i;\
  for(_i=0; _i<ftmext_get_inlets_dsp(_e); _i++) { ftmext_set_connected(_e, _i, count[_i]); } /* init inlet connected array */
  
#define FTMEXT_GET_DSP_SR() (_dsp_sr)
#define FTMEXT_GET_DSP_SIZE() (_dsp_size)  
#define FTMEXT_DSP_INIT_RETURN maxext_dsp_add(_e, _pr, _sp, ftmext_get_inlets_dsp(_e) + ftmext_get_outlets_dsp(_e)); _ftmext_exit_label: ; }

/* dsp performance */
#define FTMEXT_DSP_PERFORM(p) static t_int * _FTMEXT_GEN_METHOD_NAME(p, _perform) (t_int *_dsp_words) { \
  ftmext_t *_e = (ftmext_t *)(_dsp_words[1]); int _dsp_size = (int)(_dsp_words[2]);
  
#define FTMEXT_GET_DSP_INPUT(i) ((float *)(_dsp_words[3 + (i)]))
#define FTMEXT_GET_DSP_OUTPUT(i) ((float *)(_dsp_words[3 + ftmext_get_inlets_dsp(_e) + (i)]))

#define FTMEXT_DSP_PERFORM_RETURN _ftmext_exit_label: return (_dsp_words + 3 + ftmext_get_inlets_dsp(_e) + ftmext_get_outlets_dsp(_e)); }
  
#endif

/****************************************************************************
 *
 *  FTM Max/MSP external actions and handlers
 *
 */
#ifndef AVOID_MACROS
  
#define FTMEXT_CLASS_ACTION_OPEN(p) do { addmess((method) _FTMEXT_GEN_METHOD_NAME(p, handler_open), "dblclick", A_CANT, 0); } while(0)
#define FTMEXT_CLASS_ACTION_SAVE(p) do { ftmext_class_save_handler(_c, _FTMEXT_GEN_METHOD_NAME(p, handler_save)); } while(0)
#define FTMEXT_HANDLER_SAVE(p) static void _FTMEXT_GEN_METHOD_NAME(p, handler_save) (ftmext_t *_e) { 
#define FTMEXT_HANDLER_OPEN(p) static void _FTMEXT_GEN_METHOD_NAME(p, handler_open) (ftmext_t *_e) {
#define FTMEXT_HANDLER_RETURN _ftmext_exit_label: ; }
  
#endif
  
/****************************************************************************
 *
 *  FTM Max/MSP external alarm
 *
 */
#ifndef AVOID_MACROS
  
#define FTMEXT_CLASS_ALARM(p, m, d) do { ftmext_class_alarm(_c, _FTMEXT_GEN_METHOD_NAME(p, m), (d)); } while(0)
  
#endif
  
  /*********************************************************************************
 *
 *  FTM Max/MSP external helper functions
 *
 */
/* Max/MSP specific functions (defined in maxpat.c) */
FTMEXT_API t_patcher *maxpat_get_toplevel(t_patcher *pat);
FTMEXT_API int maxpat_get_depth(t_patcher *pat);
FTMEXT_API fts_patcher_t *maxpat_get_scope(t_patcher *pat);
FTMEXT_API fts_patcher_t *maxpat_get_scope_by_depth(t_patcher *pat, int depth);
FTMEXT_API char *maxpat_get_description(t_patcher *pat);
FTMEXT_API void maxpat_unresolved_scope_add(t_object *ext, t_patcher *pat);
FTMEXT_API int maxpat_has_unresolved(void);

/* Max/MSP specific symbols (defined in maxext.c) */
FTMEXT_API fts_symbol_t maxext_sym_arrow_right;
FTMEXT_API fts_symbol_t maxext_sym_arrow_left;
FTMEXT_API fts_symbol_t maxext_sym_hash_N;
FTMEXT_API fts_symbol_t maxext_sym_hash_P;
FTMEXT_API fts_symbol_t maxext_sym_hash_T;
FTMEXT_API void maxext_update_scope(ftmext_t *ext);

/*********************************************************************************
 *
 *  FTM Max/MSP external tables for graphic interaction
 *
 */
/* Max/MSP specific symbols (defined in maxext.c) */
FTMEXT_API void maxext_edithash_add(t_patcher *patcher, t_box *obj);
FTMEXT_API void maxext_edithash_remove(t_patcher *patcher, t_box *obj);
FTMEXT_API void maxext_edithash_print(t_patcher *patcher);
FTMEXT_API int maxext_edithash_contains( t_patcher *patcher);
FTMEXT_API t_box *maxext_edithash_clicked_object(t_patcher *patcher, int x, int y);
FTMEXT_API t_box *maxext_edit_get_active_box( void);
FTMEXT_API void maxext_edit_set_active_box(t_box *box);

#ifndef WIN32
FTMEXT_API t_patcher *maxext_edithash_current_patcher_get( EventRef theEvent);
#else
FTMEXT_API t_patcher *maxext_edithash_current_patcher_get(WPARAM wParam, LPARAM lParam);
#endif

/*********************************************************************************
 *
 *  DEPRECATED FTM Max/MSP external API functions
 *
 */
FTMEXT_API void ftm_outlet_object(void *outlet, fts_object_t *obj);
FTMEXT_API void ftm_outlet_atom(void *outlet, const fts_atom_t *a);
FTMEXT_API void ftm_error_input_object(fts_object_t *obj, const char *prompt);

#endif
