/* $Id: ftmsdif.h,v 1.3 2007/09/14 18:05:13 diemo Exp $
 *
 * jMax
 * Copyright (C) 1994-2002 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * See file LICENSE for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * This file's authors:  Diemo Schwarz, 20.8.2002
 */

/* 
 * $Log: ftmsdif.h,v $
 * Revision 1.3  2007/09/14 18:05:13  diemo
 * allow temp space for fvec copy and writing
 *
 * Revision 1.2  2007/04/11 16:37:34  nschnell
 *  - changed jMax header
 *  - reintroduced timebase
 *
 * Revision 1.1  2007/02/22 11:31:51  nschnell
 *  - included SDIF (formerly in jMax sdif package)
 *
 * Revision 1.24  2006/04/28 13:07:45  diemo
 * write revision of caller to nvt
 * untabify
 *
 * Revision 1.23  2006/04/21 10:17:36  borghesi
 * adapted to windows version
 *
 * Revision 1.22  2006/04/21 08:56:57  borghesi
 * methods exported for the windows version
 *
 * Revision 1.21  2006/03/09 09:03:01  diemo
 * jmaxsdif_get_time implemented on msp (should be in ftm_rt?)
 *
 * Revision 1.20  2006/01/12 18:11:50  diemo
 * fixed track export: added missing mode initialisation
 *
 * Revision 1.19  2005/12/13 08:50:05  diemo
 * - generalise import wrapper to prepare for data object sdif import (bpf, fmat)
 * - scoob track to sdif-midi export (keeps precise timing!)
 * - sketch for pseudo sdif class to report instantiation of track to later
 *
 * Revision 1.18  2005/10/21 14:48:10  diemo
 * added export/import mode field
 * keep quiet in release build
 * remove signed/unsigned warnings
 * use safe vsnprintf
 *
 * Revision 1.17  2005/05/27 15:46:49  diemo
 * added missing update of stream selection int mask
 * adapted to new sdif error tag declarations
 *
 * Revision 1.16  2005/05/13 16:25:14  diemo
 * non-crashing when importing corrupt sdif files (\n -> \r\n)
 *
 * Revision 1.15  2005/03/02 13:56:42  diemo
 * declare jmaxsdif_set_file
 *
 * Revision 1.14  2005/02/02 15:25:59  diemo
 * max' post crashes in any case???
 * all prototypes
 * as soon as a file is sdif, we handle it, even on error
 *
 * Revision 1.13  2005/01/14 18:16:19  diemo
 * jmaxsdif_error does the right thing, replaces fts_object_error!
 * don't use fts_get_string!
 * leave out externals for the moment
 * test all
 *
 * Revision 1.12  2004/10/12 16:01:32  diemo
 * use export handler
 * filename parsing done by SdifGetFilenameAndSelection first,
 * to remove selection, no need to strdup filename (symbols live for ever)
 * ftmext.h for jmax needed to compile jmaxsdif
 *
 * Revision 1.11  2004/10/07 17:49:13  diemo
 * compiles for msp on macho
 *
 * Revision 1.10  2004/09/10 15:30:20  diemo
 * sdifinfo object reads an SDIF file and prints info about the data in it or
 * copies it to an atom matrix if given
 *
 * Revision 1.9  2004/09/07 16:29:42  diemo
 * fixed new and old double free bugs (thanks to Patrice and valgrind)
 * by using SdifSelectAppendList, avoids double references --> import ok
 *
 * Revision 1.8  2004/09/06 17:13:30  diemo
 * jmaxsdif_init_from_object
 * export ok
 *
 * Revision 1.7  2004/09/06 13:28:16  diemo
 * First sketch of exporting.
 * Moved more common stuff to jmaxsdif.c
 *
 * Revision 1.6  2004/09/03 13:40:50  diemo
 * simple track import working
 * make compile with new SDIF using pthreads
 *
 * Revision 1.5  2004/01/09 15:51:49  diemo
 * remove compatibility for jmax 4.0.2,
 * now we're 0.7 for jmax 4.1
 *
 * Revision 1.4  2004/01/09 15:38:33  diemo
 * temporary compatibility for jmax 4.0.2
 * compile warning-free
 *
 * Revision 1.3  2003/07/08 19:15:19  diemo
 * Typecheck, avoid writing files without defining the types
 * (this should be done by the sdif library)
 * I/O of symbols in text matrices.
 *
 * Revision 1.2  2003/02/11 11:10:12  p_tisserand
 * Fix compilation with jMax release 4.0.1
 *
 * Revision 1.1.1.1  2002/12/19 19:25:44  diemo
 * Import of jmax package sdif, version 0.5.
 *
 * Revision 1.5  2002/12/13 17:14:29  schwarz
 * Move to jMax 4 API.
 * Changes in instantiate/init, tuple/varargs distinction, etc.
 *
 * Revision 1.4  2002/09/27 09:44:00  schwarz
 * Full matrix spec in framespec.
 * Streamid default as parameter (all streams/stream 0).
 * Common function jmaxsdif_openfile.
 *
 * Revision 1.3  2002/09/18 10:45:36  schwarz
 * New init argument parser works (but is not used, yet).
 * Init/delete methods in "base class" jmaxsdif.
 * Proper error reporting with fts_object_set_error.
 * VERSION now from configure.in.
 *
 * Revision 1.2  2002/09/10 14:11:02  schwarz
 * Added autogen.sh
 *
 * Revision 1.1.1.1  2002/09/10 09:31:40  schwarz
 * Moved package sdif from jmax source tree to separate cvs module.
 *
 * $Log  jmaxsdif.h,v  from old module jmax/packages/sdif: $
 * Revision 1.2  2002/09/02 10:39:14  schwarz
 * Working list i/o.
 * Version output included.
 */


#ifndef _JMAXSDIF_H_
#define _JMAXSDIF_H_

#ifdef WIN32
#if defined(FTM_EXPORTS)
#define JMAXSDIF_API __declspec(dllexport)
#else
#define JMAXSDIF_API __declspec(dllimport)
#endif
#else
#define JMAXSDIF_API extern
#endif

#include "fts.h"
#include "ftmext.h"
#include <sdif.h>


/* maximum number of scalar or list data elements for one frame */
#define MAX_NUM         1024

/* max length of string matrix */
#define MAX_TEXT        16000


typedef enum { state_stopped, state_open, state_running } jmaxsdif_state_t;

typedef enum { mode_import, mode_export } jmaxsdif_mode_t;


/* frame/matrix structures for argument parsing and data i/o: */

/* parsed matrix specification */
typedef struct
{
    int                 present;
    SdifListT*          matrixsig;      /* matrix type(s) to write/(read) */
    int                 nrows;          /* for writesdif */
    int                 ncols;
    float*              data;
    SdifDataTypeET      type;
    fmat_t*             fmat;           /* for readsdif */
    float*              temp;		/* space for temporary data */
    /* TODO: 
       enum  { text, real, int }  type;
       union { fts_symbol_t text; fmat_t* fmat; imat_t *imat; } */
} matrixspec_t;


/* parsed frame specification from object init args:
   describes one frame and m matrices in it to read or write */
typedef struct
{
    int                 present;
    unsigned int        streamid;       /* TODO: list of SdifSelectElementT */
    SdifListT*          framesig;       /* list of SdifSelectElementT */
    int                 nmatrix;
    matrixspec_t*       matrixspec;     /* array [nmatrix] of matrixspec_t */
} framespec_t;


/* object base class */
typedef struct
{
    ftmext_t            o;              /* mandatory ftmext base class */
    fts_object_t *      object;         /* if not NULL, our real object is here */
    const char*         objectname;     /* who are we? */

    jmaxsdif_mode_t     mode;           /* import or export */
    jmaxsdif_state_t    state;
    double              starttime;      /* time of start message */

    int                 nframes;
    framespec_t*        framespec;      /* array [nframes] of framespec_t */
    SdifUInt4           defaultstreamid;

    const char*         filename;       /* filename::selection argument */
    SdifSelectionT      selection;      /* holds parsed filename */
    SdifFileT*          file;           /* NULL if not open */
} jmaxsdif_t;


/* base class of data available to import callbacks */
typedef struct
{
    jmaxsdif_t    base;   /* only used to store frame specs and file name */
    fts_object_t *target;
    int           selrow; /* first selected row/column */
    int           selcol;
} jmaxsdif_import_data_t;


extern fts_symbol_t readsdif_symbol;
extern fts_symbol_t writesdif_symbol;
extern fts_symbol_t s_define;
extern fts_symbol_t s_open;
extern fts_symbol_t s_record;
extern fts_symbol_t s_close;


/*
 * prototypes 
 */

JMAXSDIF_API double jmaxsdif_get_time (void);

/* package init */
void sdif_config(void);
void seqsdif_config (void);

/* class init */

/** exit function: called when the sdif lib wants to exit */
void jmaxsdif_exit(void);

/** handler for sdif library errors and warnings */
void jmaxsdif_errorhandler(SdifErrorTagET   error_tag, 
                           SdifErrorLevelET error_level, 
                           char*            error_message, 
                           SdifFileT*       error_file, 
                           SdifErrorT*      error_ptr, 
                           char*            source_file, 
                           int              source_line);

/* jmaxsdif base class methods */

/** init jmaxsdif object, count framespecs */
JMAXSDIF_API int jmaxsdif_init(jmaxsdif_t *base, int ac, const fts_atom_t* at);

/** init jmaxsdif object from another fts object, count framespecs, 
    parse filename and selection */
JMAXSDIF_API int jmaxsdif_init_from_object (jmaxsdif_t *base, fts_object_t *o, 
                               char *objname, jmaxsdif_mode_t mode,
                               int ac, const fts_atom_t* at);

/** delete object, free memory */
JMAXSDIF_API void jmaxsdif_delete(jmaxsdif_t *base);

/** error reporting that calls the appropriate function */
JMAXSDIF_API void jmaxsdif_error(jmaxsdif_t *base, const char *format, ...);

/** set file name in base, handle path normalisation */
JMAXSDIF_API void jmaxsdif_set_file (jmaxsdif_t *base, const fts_atom_t* at);

/** open file in base
    return: success */
JMAXSDIF_API int  jmaxsdif_openfile(jmaxsdif_t *base, SdifFileModeET mode);

/** open file in base and set up NVT
 *  @return: success */
JMAXSDIF_API int jmaxsdif_openfile_write(jmaxsdif_t *base, 
					 const char *writtenby, 
                                         const char *revision);

/** conditional close of file in base->file
    return: success */
JMAXSDIF_API int  jmaxsdif_closefile(jmaxsdif_t *base);

/** just count the frames in the args, don't parse on instantiate */
JMAXSDIF_API int  jmaxsdif_countframes(int ac, const fts_atom_t* at);

/** framespec parser
    @return number of consumed atoms, sets base->nframes */
JMAXSDIF_API int  jmaxsdif_parseargs(jmaxsdif_t *base, int ac, const fts_atom_t* at,
                        SdifUInt4 defaultid);

/** add sdif type definition given in strings to open file */
JMAXSDIF_API int jmaxsdif_typedef (jmaxsdif_t *base, int ac, const fts_atom_t *at);

/** check if all frame/matrix signatures in the framespecs are in
    defined SDIF types */
JMAXSDIF_API int  jmaxsdif_typecheck(jmaxsdif_t *base);

/** set matrix spec to zero */
void jmaxsdif_clearmatrix(matrixspec_t* mtx);

void jmaxsdif_framespec_print(jmaxsdif_t *base);

/** general import function, loops over sdif file, callbacks to create data */
JMAXSDIF_API int jmaxsdif_import (jmaxsdif_import_data_t *import, 
                     int ac, const fts_atom_t* at,
                     SdifOpenFileCallbackT   openfilefunc,
                     SdifFrameCallbackT      framefunc,
                     SdifMatrixDataCallbackT matrixdatafunc,
                     /*out*/ fts_atom_t *ret);

/* object init */
/*
FTM_INSTANTIATE(readsdif_instantiate);
FTM_INSTANTIATE(writesdif_instantiate);
*/
void sdifinfo_config(void);


#endif /* ifndef _JMAXSDIF_H_ */
