/*
 * FTM
 * Copyright (C) 2004 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#ifndef _FTS_PATCHER_OBJECT_H_
#define _FTS_PATCHER_OBJECT_H_

#include "fts.h"

/* open/save dialog */
#define fts_object_open_dialog ftmrte_file_open_panel
#define fts_object_save_dialog ftmrte_file_save_panel

/* object error */
FTS_API void fts_object_error(fts_object_t *obj, const char *format, ...);

/* id string (name, client id, pointer address) */
FTS_API char *fts_object_get_identifier_string(fts_object_t *obj, char *str, int len);
FTS_API char *fts_object_get_short_identifier_string(fts_object_t *obj, char *str, int len);

/* for post.c */
#define fts_object_get_description_size(o) (0)
#define fts_object_get_description_atoms(o) (NULL)

/***************************************************************************
 *
 *  FTM persistent object and name definiton
 *
 */

struct ftm_context
{
  fts_object_t *container; /* container replacing patcher */
  ftmrte_external_t *ext; /* back pointer to external */
  fts_string_t description; /* raw description */
  fts_memorystream_t *persistent_display_string; /* displayed description */
  fts_array_t args; /* tokenized description */
  fts_expression_t *expression; /* parsed expression */
  fts_status_t status;
  char error[256];
  fts_atom_t value;
  fts_object_t *persistent_reference;
  fts_symbol_t name;
  int global;
  fts_definition_t *definition;
  int persistence;
  int can_dirty;
};

typedef struct ftm_context ftm_context_t;

#define ftm_context_get_reference(c) ((c)->persistent_reference)

/* context */
FTS_API void ftm_context_init(ftm_context_t *context, ftmrte_external_t *self);
FTS_API void ftm_context_delete(ftm_context_t *context);

FTS_API fts_status_t ftm_context_set_from_args(ftm_context_t *context, int ac, const fts_atom_t *at);
FTS_API ftm_context_t *ftm_context_get(fts_object_t *obj);

/* description */
FTS_API void ftm_context_set_description(ftm_context_t *context, char *string);
FTS_API char *ftm_context_get_description(ftm_context_t *context);
FTS_API const char *ftm_context_get_error(ftm_context_t *context);

/* persistence */
FTS_API void ftm_context_set_persistence(ftm_context_t *context, int persistence);
FTS_API int ftm_context_is_persistent(ftm_context_t *context);
FTS_API void ftm_context_set_dirty(ftm_context_t *context);
FTS_API void ftm_context_set_state_dirty(ftm_context_t *context);
FTS_API void ftm_context_can_dirty(ftm_context_t *context, int can);

/* value */
FTS_API void ftm_context_set_value(ftm_context_t *context, const fts_atom_t *value);
FTS_API fts_atom_t *ftm_context_get_value(ftm_context_t *context);

/* name and scope */
FTS_API void ftm_context_set_name(ftm_context_t *context, fts_symbol_t wanted_name, int global);
FTS_API fts_symbol_t ftm_context_get_name(ftm_context_t *context);
FTS_API int ftm_context_is_global(ftm_context_t *context);
FTS_API char *ftm_context_get_scope_description(ftm_context_t *context);
FTS_API void ftm_context_set_scope(ftm_context_t *context, int global);

/* context dump */
FTS_API void ftm_context_dump(ftm_context_t *context, void *data);
FTS_API void ftm_context_dump_object(ftm_context_t *context);

/* context restore utils */
FTS_API void ftm_context_restore_scope(ftm_context_t *context, int ac, fts_atom_t *at);
FTS_API void ftm_context_restore_object(ftm_context_t *context, int ac, fts_atom_t *at);
FTS_API void ftm_context_restore_message(ftm_context_t *context, int ac, fts_atom_t *at);
FTS_API void ftm_context_restore_definition(ftm_context_t *context, int ac, fts_atom_t *at);

FTS_API void ftm_context_redefine(ftm_context_t *context);

/* upload */
FTS_API void ftm_open_editor(fts_object_t *obj);
FTS_API void ftm_close_editor(fts_object_t *obj);
FTS_API int ftm_has_editor(fts_object_t *obj);

FTS_API void ftm_context_open_editor(ftm_context_t *context);
FTS_API void ftm_context_close_editor(ftm_context_t *context);

FTS_API void ftm_context_upload_name(ftm_context_t *context);
FTS_API void ftm_context_upload_object(ftm_context_t *context);

/***************************************************************************
 *
 *  FTM dumper & loader
 *
 */
FTS_API void ftm_dumper_begin(void *data, fts_patcher_t *scope);
FTS_API void ftm_dumper_end(void);
FTS_API void ftm_dumper_reset(void);
FTS_API void *ftm_dumper_get_handle(void);
FTS_API void ftm_dumper_save_message(fts_symbol_t selector, int ac, const fts_atom_t *at);

FTS_API void ftm_loader_reset(void);
FTS_API void ftm_loader_restore_ftm(int ac, fts_atom_t *at);
FTS_API void ftm_loader_restore_scope(int ac, fts_atom_t *at);
FTS_API void ftm_loader_restore_absargs(int ac, fts_atom_t *at);
FTS_API void ftm_loader_save_absargs(fts_array_t *array);
FTS_API int ftm_loader_string_replace_absargs(char *str, char *buf, int len);
FTS_API void ftm_loader_replace_args(int ac, const fts_atom_t *at, fts_atom_t *ret_at);

FTS_API char ftm_str_restore_ftm[];
FTS_API char ftm_str_restore_scope[];
FTS_API char ftm_str_restore_obj[];
FTS_API char ftm_str_restore_mess[];
FTS_API char ftm_str_restore_def[];
FTS_API char ftm_str_restore_absargs[];

FTS_API fts_symbol_t ftm_sym_restore_scope;
FTS_API fts_symbol_t ftm_sym_restore_obj;
FTS_API fts_symbol_t ftm_sym_restore_mess;
FTS_API fts_symbol_t ftm_sym_restore_def;
FTS_API fts_symbol_t ftm_sym_restore_absargs;

/* utility function */
FTS_API fts_symbol_t ftm_symbol_replace_hashes(fts_symbol_t sym, int ac, fts_atom_t *at);

#endif
